import pytest
import os
import sys
import click
from click.testing import CliRunner
import git
import requests
import json
from wandb import util
from wandb.apis import InternalApi

import webbrowser
from wandb.git_repo import GitRepo
from distutils.version import LooseVersion

torch = util.get_module("torch")
if torch:
    if LooseVersion(torch.__version__) < LooseVersion("0.4"):
        pytorch_tensor = torch.Tensor
        OLD_PYTORCH = True
    else:
        # supports 0d tensors but is a module before 0.4
        pytorch_tensor = torch.tensor
        OLD_PYTORCH = False


@pytest.fixture
def runner(monkeypatch, mocker):
    whaaaaat = util.vendor_import("whaaaaat")
    monkeypatch.setattr('wandb.cli.api', InternalApi(
        default_settings={'project': 'test', 'git_tag': True}, load_settings=False))
    monkeypatch.setattr(click, 'launch', lambda x: 1)
    monkeypatch.setattr(whaaaaat, 'prompt', lambda x: {
                        'project_name': 'test_model', 'files': ['weights.h5'], 'attach': False,
                        'team_name': 'Manual Entry'})
    monkeypatch.setattr(webbrowser, 'open_new_tab', lambda x: True)
    return CliRunner()


@pytest.fixture
def git_repo():
    with CliRunner().isolated_filesystem():
        r = git.Repo.init(".")
        os.mkdir("wandb")
        # Because the forked process doesn't use my monkey patch above
        with open("wandb/settings", "w") as f:
            f.write("[default]\nproject: test")
        open("README", "wb").close()
        r.index.add(["README"])
        r.index.commit("Initial commit")
        yield GitRepo(lazy=False)


def assert_deep_lists_equal(a, b, indices=None):
    try:
        assert a == b
    except ValueError:
        assert len(a) == len(b)

        # pytest's list diffing breaks at 4d so we track them ourselves
        if indices is None:
            indices = []
            top = True
        else:
            top = False

        for i, (x, y) in enumerate(zip(a, b)):
            try:
                assert_deep_lists_equal(x, y, indices)
            except AssertionError:
                indices.append(i)
                raise
            finally:
                if top and indices:
                    print('Diff at index: %s' % list(reversed(indices)))


def subdict(d, expected_dict):
    """Return a new dict with only the items from `d` whose keys occur in `expected_dict`.
    """
    return {k: v for k, v in d.items() if k in expected_dict}


class ResponseMock(object):
    def __init__(self, response):
        self.response = response

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def raise_for_status(self):
        if self.response.status_code >= 400:
            raise requests.exceptions.HTTPError("Bad Request", response=self.response)

    @property
    def content(self):
        return self.response.data.decode('utf-8')

    def iter_content(self, chunk_size=1024):
        yield self.response.data

    def json(self):
        return json.loads(self.response.data.decode('utf-8'))


class RequestsMock(object):
    def __init__(self, app, ctx):
        self.app = app
        self.client = app.test_client()
        self.ctx = ctx

    def set_context(self, key, value):
        self.ctx[key] = value

    def Session(self):
        return self

    @property
    def RequestException(self):
        return requests.RequestException

    @property
    def HTTPError(self):
        return requests.HTTPError

    @property
    def headers(self):
        return {}

    @property
    def utils(self):
        return requests.utils

    @property
    def exceptions(self):
        return requests.exceptions

    @property
    def packages(self):
        return requests.packages

    @property
    def adapters(self):
        return requests.adapters

    def mount(self, *args):
        pass

    def _clean_kwargs(self, kwargs):
        if "auth" in kwargs:
            del kwargs["auth"]
        if "timeout" in kwargs:
            del kwargs["timeout"]
        if "cookies" in kwargs:
            del kwargs["cookies"]
        if "params" in kwargs:
            del kwargs["params"]
        if "stream" in kwargs:
            del kwargs["stream"]
        if "verify" in kwargs:
            del kwargs["verify"]
        if "allow_redirects" in kwargs:
            del kwargs["allow_redirects"]
        return kwargs

    def _store_request(self, url, body):
        key = url.split("/")[-1]
        self.ctx[key] = self.ctx.get(key, [])
        self.ctx[key].append(body)

    def post(self, url, **kwargs):
        self._store_request(url, kwargs.get("json"))
        return ResponseMock(self.client.post(url, **self._clean_kwargs(kwargs)))

    def put(self, url, **kwargs):
        self._store_request(url, kwargs.get("json"))
        return ResponseMock(self.client.put(url, **self._clean_kwargs(kwargs)))

    def get(self, url, **kwargs):
        self._store_request(url, kwargs.get("json"))
        return ResponseMock(self.client.get(url, **self._clean_kwargs(kwargs)))

    def request(self, method, url, **kwargs):
        if method.lower() == "get":
            self.get(url, **kwargs)
        elif method.lower() == "post":
            self.post(url, **kwargs)
        elif method.lower() == "put":
            self.put(url, **kwargs)
        else:
            raise requests.RequestException("Request method not implemented: %s" % method)


import matplotlib.pyplot as plt
plt.switch_backend('agg')

import sys
import os
import pytest
import glob
import json
import wandb
import numpy as np
import pytest
import tensorflow as tf
from tensorboardX import SummaryWriter

# Tests which rely on row history in memory should set `History.keep_rows = True`
from wandb.history import History
History.keep_rows = True


@pytest.mark.skipif(sys.version_info < (3, 6), reason="This test doesn't work in py2 tensorboard")
def test_tensorboard(run_manager):
    wandb.tensorboard.patch(tensorboardX=False, pytorch=False)
    tf.summary.FileWriterCache.clear()
    summary = tf.summary.FileWriter(".")
    summary.add_summary(tf.Summary(
        value=[tf.Summary.Value(tag="foo", simple_value=1)]), 0)
    summary.flush()
    run_manager.test_shutdown()
    print("Run History Rows", wandb.run.history.rows)
    assert wandb.run.history.rows[0]["foo"] == 1.0
    assert wandb.run.history.rows[0]["global_step"] == 0
    assert len(glob.glob(wandb.run.dir + "/*tfevents*")) == 1


@pytest.mark.skipif(sys.version_info < (3, 6), reason="This test doesn't work in py2 tensorboard")
def test_tensorboard_no_step(run_manager):
    wandb.tensorboard.patch(tensorboardX=False, pytorch=False)
    tf.summary.FileWriterCache.clear()
    summary = tf.summary.FileWriter(".")
    summary.add_summary(tf.Summary(
        value=[tf.Summary.Value(tag="foo", simple_value=1)]), 0)
    wandb.log({"foo": 10, "bar": 32})
    summary.add_summary(tf.Summary(
        value=[tf.Summary.Value(tag="foo", simple_value=2)]), 1)
    summary.flush()
    run_manager.test_shutdown()
    assert wandb.run.history.rows[1]["foo"] == 2
    assert wandb.run.history.rows[0]["bar"] == 32
    assert len(wandb.run.history.rows) == 2


@pytest.mark.mocked_run_manager()
def test_tensorboard_load_complex(wandb_init_run):
    """This test is to ensure the final event logged in a given step remains in that step"""
    steps_for_meta = []
    for summary in tf.train.summary_iterator(os.path.join(os.path.dirname(__file__),
                                                          "fixtures/events.out.tfevents.111.complex.localdomain")):
        parsed = wandb.tensorboard.tf_summary_to_dict(summary)
        if parsed is None:
            continue
        if parsed.get("meta/activation/D/mean/high"):
            steps_for_meta.append(summary.step)
        wandb.tensorboard.log(summary, step=summary.step)
    wandb_init_run.run_manager.test_shutdown()
    rows = [json.loads(row) for row in open(os.path.join(wandb_init_run.dir, "wandb-history.jsonl")).readlines()]
    assert steps_for_meta == [row["global_step"] for row in rows if row.get("meta/activation/D/mean/high")]
    assert len(rows) == 43


@pytest.mark.mocked_run_manager()
def test_tensorboard_load_rate_limit_filter(wandb_init_run):
    """This test is to ensure the final event logged in a given step remains in that step"""
    try:
        wandb.tensorboard.configure(rate_limit_seconds=10, ignore_kinds=["histo"])
        for summary in tf.train.summary_iterator(os.path.join(os.path.dirname(__file__),
                                                              "fixtures/events.out.tfevents.111.complex.localdomain")):
            wandb.tensorboard.log(summary, step=summary.step)
        wandb_init_run.run_manager.test_shutdown()
        rows = [json.loads(row) for row in open(os.path.join(wandb_init_run.dir, "wandb-history.jsonl")).readlines()]
        assert len(rows) == 4
        assert rows[0].get("gradient/netG_A/norm/histogram") == None
    finally:
        wandb.tensorboard.configure()


@pytest.mark.skipif(sys.version_info < (3, 6), reason="This test doesn't work in py2 tensorboard")
def test_tensorboard_s3(run_manager, capsys, mocker):
    # This mocks out the tensorboard writer so we dont attempt to talk to s3
    from tensorflow.python.summary.writer import event_file_writer
    #from tensorboard.summary.writer import event_file_writer

    def fake_init(self, logdir, **kwargs):
        self._closed=False
        writer=event_file_writer.EventFileWriter("test")
        mocker.patch.object(writer._ev_writer, "FileName",
                            lambda: logdir.encode("utf8"))
        mocker.patch.object(writer._ev_writer, "Flush")
        super(tf.summary.FileWriter, self).__init__(writer, None, None)
    mocker.patch("tensorflow.summary.FileWriter.__init__", fake_init)
    wandb.tensorboard.patch(tensorboardX=False, pytorch=False)
    tf.summary.FileWriterCache.clear()
    summary=tf.summary.FileWriter("s3://simple/test")
    summary.add_summary(tf.Summary(
        value=[tf.Summary.Value(tag="foo", simple_value=1)]), 0)
    summary.flush()
    run_manager.test_shutdown()
    out, err=capsys.readouterr()
    print("OUT", out)
    print("ERR", err)
    assert "s3://simple/test" in err
    assert "can't save file to wandb" in err
    print(wandb.run.history.row)
    print(wandb.run.history.rows)
    assert wandb.run.history.rows[0]["foo"] == 1.0
    assert wandb.run.history.rows[0]["global_step"] == 0
    assert len(glob.glob(wandb.run.dir + "/*tfevents*")) == 0


@pytest.mark.skipif(sys.version_info < (3, 6), reason="No moviepy.editor in py2")
def test_tensorboardX(run_manager):
    wandb.tensorboard.patch(tensorboardX=True)

    fig=plt.figure()
    c1=plt.Circle((0.2, 0.5), 0.2, color='r')

    ax=plt.gca()
    ax.add_patch(c1)
    plt.axis('scaled')

    writer=SummaryWriter()
    writer.add_figure('matplotlib', fig, 0)
    writer.add_video('video', np.random.random(size=(1, 5, 3, 28, 28)), 0)
    writer.add_scalars('data/scalar_group', {
        'foo': 10,
        'bar': 100
    }, 1)
    writer.close()
    run_manager.test_shutdown()
    rows=run_manager.run.history.rows
    events=[]
    for root, dirs, files in os.walk(run_manager.run.dir):
        print("ROOT", root, files)
        for file in files:
            if "tfevent" in file:
                events.append(file)
    assert rows[0]["matplotlib"]['width'] == 640
    assert rows[0]["matplotlib"]['height'] == 480
    assert rows[0]["matplotlib"]['_type'] == 'images'
    assert rows[0]["video"]['_type'] == 'videos'
    assert rows[1]["data/scalar_group/foo"] == 10
    assert rows[1]["data/scalar_group/foo/global_step"] == 1
    assert rows[1]['data/scalar_group/bar'] == 100
    assert rows[1]['global_step'] == 1
    assert len(events) == 3



import matplotlib
# this needs to be called before importing wandb Graph
matplotlib.use("Agg")


from wandb import wandb_run
from wandb.summary import FileSummary
import pandas
import tensorflow as tf
import torch
import json
import glob
import os
import sys
import numpy as np
import tempfile
import plotly.graph_objs as go
import matplotlib.pyplot as plt
from wandb import Histogram, Image, Graph, Table
from wandb.data_types import Node
from click.testing import CliRunner
import pytest


@pytest.fixture
def summary():
    with CliRunner().isolated_filesystem():
        run = wandb_run.Run()
        run.summary.update({"foo": "init"})
        yield run.summary


def disk_summary(summary):
    return json.load(open(summary._fname))


def test_set_attrs(summary):
    summary.foo = "bar"
    assert disk_summary(summary) == {"foo": "bar"}


def test_get_attr(summary):
    assert summary.foo == "init"


def test_update(summary):
    summary.update({"foo": "bar"})
    assert disk_summary(summary) == {"foo": "bar"}


def test_update_histogram(summary):
    summary.update({"hist": Histogram(np_histogram=([1, 2, 3], [1, 2, 3, 4]))})
    assert disk_summary(summary) == {
        'foo': 'init',
        "hist": {"_type": "histogram", "values": [1, 2, 3], "bins": [1, 2, 3, 4]}}


def test_set_histogram(summary):
    summary["hist"] = Histogram(np_histogram=([1, 2, 3], [1, 2, 3, 4]))
    assert disk_summary(summary) == {
        'foo': 'init',
        "hist": {"_type": "histogram", "values": [1, 2, 3], "bins": [1, 2, 3, 4]}}


def test_set_item(summary):
    summary["foo"] = "bar"
    assert disk_summary(summary) == {"foo": "bar"}


def test_get_item(summary):
    assert summary["foo"] == "init"


def test_delete(summary):
    summary.update({"foo": "bar", "bad": True})
    assert summary['foo'] == 'bar'
    assert summary['bad'] is True
    del summary["bad"]
    assert disk_summary(summary) == {"foo": "bar"}


def test_image(summary):
    summary["image"] = Image(np.zeros((28, 28)))
    ds = disk_summary(summary)
    assert os.path.exists(os.path.join(summary._run.dir, ds['image']['path']))

    expected = {
        '_type': 'image-file',
        'height': 28,
        'width': 28,
        'size': 73,
    }
    assert set(ds['image'].items()) >= set(expected.items())


def test_matplot_image(summary):
    img = plt.imshow(np.zeros((28, 28)), cmap='gray')
    summary["fig"] = img
    plt.close()
    ds = disk_summary(summary)
    assert os.path.exists(os.path.join(summary._run.dir, ds['fig']['path']))
    assert set(ds["fig"].items()) >= set({
        "_type": "image-file",
        "height": 480,
        "width": 640,
    }.items())


def test_matplot_plotly(summary):
    plt.cla()
    plt.plot([1, 2, 3])
    summary["plot"] = plt
    plt.close()
    plot = disk_summary(summary)["plot"]
    assert plot["_type"] == "plotly-file"


def test_plotly_plot(summary):
    scatter = go.Figure(  # plotly
        data=go.Scatter(x=[0, 1, 2]),
        layout=go.Layout(
            title=go.layout.Title(text="A Bar Chart")))
    summary["plot"] = scatter
    plot = disk_summary(summary)["plot"]
    assert plot["_type"] == "plotly-file"
    path = plot["path"]
    data = open(os.path.join(summary._run.dir, path)).read()
    plot_data = json.loads(data)
    assert plot_data["data"][0]['type'] == 'scatter'

def test_plotly_big_numpy(summary):
    N = 200
    x = np.arange(N)
    y = np.arange(0, N) / N
    scatter = go.Figure(  # plotly
        data=go.Scatter(x=[0, 1, 2]),
        layout=go.Layout(
            title=go.layout.Title(text="A Bar Chart")))
    summary["plot"] = scatter
    plot = disk_summary(summary)["plot"]
    path = plot["path"]
    data = open(os.path.join(summary._run.dir, path)).read()
    plot_data = json.loads(data)
    assert plot_data["data"][0]['type'] == 'scatter'

def test_graph(summary):
    graph = Graph()
    node_a = Node('a', 'Node A', size=(4,))
    node_b = Node('b', 'Node B', size=(16,))
    graph.add_node(node_a)
    graph.add_node(node_b)
    graph.add_edge(node_a, node_b)
    summary["graph"] = graph
    graph = disk_summary(summary)["graph"]
    path = graph["path"]
    data = open(os.path.join(summary._run.dir, path)).read()
    graph_data = json.loads(data)
    assert graph_data == {
        'edges': [['a', 'b']],
        'format': 'keras',
        'nodes': [{'id': 'a', 'name': 'Node A', 'size': [4]},
                  {'id': 'b', 'name': 'Node B', 'size': [16]}]}

def test_newline(summary):
    summary["rad \n"] = 1
    summary.update({"bad \n ": 2})
    summ = disk_summary(summary)
    assert summ["rad"] == 1
    assert summ["bad"] == 2


def test_big_numpy(summary):
    summary.update({"rad": np.random.rand(1000)})
    assert disk_summary(summary)["rad"]["max"] > 0
    assert os.path.exists(os.path.join(summary._run.dir, "wandb.h5"))


def test_big_nested_numpy(summary):
    summary.update({"rad": {"deep": np.random.rand(1000)}})
    assert disk_summary(summary)["rad"]["deep"]["max"] > 0
    summary["rad"]["deep2"] = np.random.rand(1000)
    assert disk_summary(summary)["rad"]["deep2"]["max"] > 0
    assert os.path.exists(os.path.join(summary._run.dir, "wandb.h5"))


def test_torch_tensor(summary):
    summary.update({"pytorch": torch.rand(1000, 1000)})
    assert os.path.exists(os.path.join(summary._run.dir, "wandb.h5"))
    assert disk_summary(summary)["pytorch"]["_type"] == "torch.Tensor"


def test_tensorflow_tensor(summary):
    with tf.Session().as_default():
        summary.update({"tensorflow": tf.random_normal([1000])})
    assert os.path.exists(os.path.join(summary._run.dir, "wandb.h5"))
    assert disk_summary(summary)["tensorflow"]["_type"] == "tensorflow.Tensor"


@pytest.mark.skipif(sys.version_info < (3, 6), reason="Fastparquet / llvmlite is BORKED in < Python 3.6")
def test_pandas(summary):
    summary.update({"pandas": pandas.DataFrame(
        data=np.random.rand(1000), columns=['col'])})


def test_read_numpy(summary):
    summary.update({"rad": np.random.rand(1000)})
    s = FileSummary(summary._run)
    assert len(s["rad"]) == 1000


def test_read_nested_numpy(summary):
    summary.update({"rad": {"deep": np.random.rand(1000)}})
    s = FileSummary(summary._run)
    assert len(s["rad"]["deep"]) == 1000

def test_read_nested_array(summary):
    summary["rad"] = {"deep": "dish"}
    s = FileSummary(summary._run)
    assert summary["rad"]["deep"] == "dish"


def test_read_very_nested_numpy(summary):
    # Test that even deeply nested writes are written to disk.
    summary.update(
        {"a": {"b": {"c": {"d": {"e": {"f": {"g": {"h": {"i": {}}}}}}}}}})
    summary['a']['b']['c']['d']['e']['f']['g']['h']['i']['j'] = True
    assert disk_summary(summary)[
        'a']['b']['c']['d']['e']['f']['g']['h']['i']['j'] is True


def test_key_locking(summary):
    summary.update({'a': 'a'})
    assert summary['a'] == 'a'
    summary.update({'a': 'b'})
    assert summary['a'] == 'b'
    summary.update({'a': 'c'}, overwrite=False)
    assert summary['a'] == 'b'



import wandb
import lightgbm as lgb


def test_types(**kwargs):
    test_passed = True
    for k, v in kwargs.items():
        if k == 'model':
            if not (isinstance(v, lgb.basic.Booster)):
                test_passed = False
                wandb.termerror("%s is not a LightGBM model. Please try again." % (k))

    return test_passed


'''W&B callback for lightgbm

Really simple callback to get logging for each tree

Example usage:

param_list = [("eta", 0.08), ("max_depth", 6), ("subsample", 0.8), ("colsample_bytree", 0.8), ("alpha", 8), ("num_class", 10)]
config.update(dict(param_list))
lgb = lgb.train(param_list, d_train, callbacks=[wandb_callback()])
'''

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
from .utils import test_types

import lightgbm
import wandb


def wandb_callback():
    def callback(env):
        eval_results = {}
        recorder = lightgbm.record_evaluation(eval_results)
        recorder(env)

        for validation_key in eval_results.keys():
            for key in eval_results[validation_key].keys():
                wandb.log({
                    validation_key + "_" + key: eval_results[validation_key][key][0]
                }, commit=False)
        # Previous log statements use commit=False. This commits them.
        wandb.log({})
    return callback


def plot_feature_importances(model=None, feature_names=None,
                             title='Feature Importance', max_num_features=50):
    """
    Evaluates & plots the importance of each feature for the gbm fitting tasks.

    Should only be called with a fitted model (otherwise an error is thrown).
    Only works with LightGBM model that have a `feature_importance` attribute.

    Arguments:
        model: Takes in a fitted gbm i.e. LightGBM
        feature_names (list): Names for features. Makes plots easier to read by
                                replacing feature indexes with corresponding names.

    Returns:
        Nothing. To see plots, go to your W&B run page then expand the 'media' tab
              under 'auto visualizations'.

    Example:
        wandb.lightgbm.plot_feature_importances(model, ['width', 'height, 'length'])
    """
    attributes_to_check = 'feature_importance'

    if not hasattr(model, attributes_to_check):
        wandb.termwarn(
            "%s attribute not in model. Cannot plot feature importances." % attributes_to_check)
        return

    if (test_types(model=model)):
        feature_names = np.asarray(feature_names)
        importances = model.feature_importance()

        indices = np.argsort(importances)[::-1]

        if feature_names is None:
            feature_names = indices
        else:
            feature_names = np.array(feature_names)[indices]

        max_num_features = min(max_num_features, len(importances))

        feature_names = feature_names[:max_num_features]
        importances = importances[:max_num_features]

        # Draw a stem plot with the influence for each instance
        # format:
        # x = feature_names[:max_num_features]
        # y = importances[indices][:max_num_features]
        def feature_importances_table(feature_names, importances):
            return wandb.visualize(
                'wandb/feature_importances/v1', wandb.Table(
                    columns=['feature_names', 'importances'],
                    data=[
                        [feature_names[i], importances[i]] for i in range(len(feature_names))
                    ]
                ))

        wandb.log({'feature_importances': feature_importances_table(feature_names, importances)})
        print("Go to your dashboard to see the LightGBM's Feature Importances graph plotted.")
        return feature_importances_table(feature_names, importances)



import pytest
from sklearn.naive_bayes import MultinomialNB
from sklearn.linear_model import ElasticNet
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.datasets import make_regression, make_hastie_10_2
from keras.layers import Dense, Flatten, Reshape
from keras.models import Sequential
from keras import backend as K
from wandb.keras import WandbCallback
import wandb
from wandb.sklearn import learning_curve, roc, confusion_matrix, precision_recall, plot_feature_importances

@pytest.fixture
def dummy_classifier(request):
    nb = MultinomialNB()
    x_train = [[1,2],[1,2],[1,2],[1,2],[2,3],[3,4],[3,4],[3,4],[3,4],[3,4],[3,4]]
    y_train = [0,0,0,0,0,1,1,1,1,1,1]
    nb.fit(x_train, y_train)
    x_test = [[4,5], [5,6]]
    y_test = [0,1]
    y_probas = nb.predict_proba(x_test)
    y_pred = nb.predict(x_test)
    return (nb, x_train, y_train, x_test, y_test, y_pred, y_probas)

@pytest.fixture
def dummy_model(request):
    K.clear_session()
    multi = request.node.get_closest_marker('multiclass')
    image_output = request.node.get_closest_marker('image_output')
    if multi:
        nodes = 10
        loss = 'categorical_crossentropy'
    else:
        nodes = 1
        loss = 'binary_crossentropy'
    nodes = 1 if not multi else 10
    if image_output:
        nodes = 300
    model = Sequential()
    model.add(Flatten(input_shape=(10, 10, 3)))
    model.add(Dense(nodes, activation='sigmoid'))
    if image_output:
        model.add(Dense(nodes, activation="relu"))
        model.add(Reshape((10, 10, 3)))
    model.compile(optimizer='adam',
                  loss=loss,
                  metrics=['accuracy'])
    return model

@pytest.fixture
def dummy_data(request):
    multi = request.node.get_closest_marker('multiclass')
    image_output = request.node.get_closest_marker('image_output')
    cats = 10 if multi else 1
    import numpy as np
    data = np.random.randint(255, size=(100, 10, 10, 3))
    labels = np.random.randint(2, size=(100, cats))
    if image_output:
        labels = data
    return (data, labels)

def test_learning_curve(dummy_classifier):
    (nb, x_train, y_train, x_test, y_test, y_pred, y_probas) = dummy_classifier
    lc_table = learning_curve(nb, x_train, y_train)
    assert(len(lc_table.value.data) == 10)
    assert(lc_table.value.data[0][0] == 'train')
    assert(lc_table.value.data[1][0] == 'test')

def test_roc(dummy_classifier):
    (nb, x_train, y_train, x_test, y_test, y_pred, y_probas) = dummy_classifier
    lc_table = learning_curve(nb, x_train, y_train)
    r = roc(y_test, y_probas)

    assert(r.value.data[0] == [0, 0.0, 0.0])

def test_confusion_matrix(dummy_classifier):
    (nb, x_train, y_train, x_test, y_test, y_pred, y_probas) = dummy_classifier
    cm = confusion_matrix(y_test, y_pred)
    assert(len(cm.value.data)==4)
    assert(cm.value.data[0]== [0,0,0])

def test_precision_recall(dummy_classifier):
    (nb, x_train, y_train, x_test, y_test, y_pred, y_probas) = dummy_classifier
    pr = precision_recall(y_test, y_probas)

    assert(pr.value.data[0]== [0, 1.0, 1.0])

def test_feature_importance_attribute_does_not_exists(wandb_init_run, dummy_model, dummy_data):
    dummy_model.fit(*dummy_data, epochs=2, batch_size=36,
                    callbacks=[WandbCallback()])
    dummy_features = []

    result = plot_feature_importances(dummy_model, feature_names=dummy_features)

    assert result is None

def test_feature_importance_attribute_exists_for_elasticnet(wandb_init_run):
    X, y = make_regression(n_features=2, random_state=42)
    two_features = ['a', 'b']
    model = ElasticNet(random_state=42)
    model.fit(X, y)

    result = plot_feature_importances(model, feature_names=two_features)

    assert isinstance(result, wandb.viz.Visualize)

def test_feature_importance_attribute_exists_for_random_forest(wandb_init_run):
    X, y = make_hastie_10_2(random_state=0)

    model = GradientBoostingClassifier(n_estimators=100, learning_rate=1.0, max_depth=1, random_state=42)
    ten_features = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'k']
    model.fit(X, y)

    result = plot_feature_importances(model, feature_names=ten_features)

    assert isinstance(result, wandb.viz.Visualize)


import pytest
import os
from click.testing import CliRunner

from wandb import wandb_config as config
from wandb import env
import yaml
import sys
import six
import os
import argparse
import textwrap


def test_config_empty_by_default():
    with CliRunner().isolated_filesystem():
        conf = config.Config()
        assert list(conf.keys()) == []


def test_config_docker_env():
    try:
        os.environ[env.DOCKER] = "ubuntu"
        with CliRunner().isolated_filesystem():
            conf = config.Config()
            assert conf["_wandb"]["docker_image"] == "ubuntu"
    finally:
        del os.environ[env.DOCKER]


def test_config_docker_env_digest():
    try:
        os.environ[env.DOCKER] = "ubuntu@sha25612345678901234567890"
        with CliRunner().isolated_filesystem():
            conf = config.Config()
            assert conf["_wandb"]["docker_image"] == "ubuntu"
            assert conf["_wandb"]["docker_digest"] == "sha25612345678901234567890"
    finally:
        del os.environ[env.DOCKER]


def test_config_set_items():
    with CliRunner().isolated_filesystem():
        conf = config.Config()
        conf['a'] = 15
        conf.b = 16
        assert list(conf.keys()) == ['a', 'b']
        assert conf.a == 15
        assert conf['b'] == 16


def test_config_accepts_dict_vals():
    with CliRunner().isolated_filesystem():
        conf = config.Config()
        conf['a'] = {'b': 14, 'c': 15}
        assert conf.a == {'b': 14, 'c': 15}


def test_config_update_accepts_dict_vals():
    with CliRunner().isolated_filesystem():
        conf = config.Config()
        conf.update({'b': 14, 'c': 15})
        assert conf.b == 14
        assert conf.c == 15


def test_config_static_match():
    with CliRunner().isolated_filesystem():
        conf = config.Config()
        conf.update({'b': 14, 'c': 15})
        stat = config.ConfigStatic(conf)
        assert stat.b == 14
        assert stat.c == 15
        assert set(conf.keys()) == set(['b', 'c'])
        assert set(stat.keys()) == set(['b', 'c'])


def test_config_static_frozen():
    with CliRunner().isolated_filesystem():
        conf = config.Config()
        conf.update({'b': 14, 'c': 15})
        stat = config.ConfigStatic(conf)
        conf.update({'d': 16, 'e': 17})
        assert set(conf.keys()) == set(['b', 'c', 'd', 'e'])
        assert set(stat.keys()) == set(['b', 'c'])


def test_config_static_readonly():
    with CliRunner().isolated_filesystem():
        conf = config.Config()
        conf.update({'b': 14, 'c': 15})
        stat = config.ConfigStatic(conf)
        with pytest.raises(AttributeError):
            stat.b = 3
        with pytest.raises(AttributeError):
            stat.new_attrib = 9


@pytest.mark.skipif(os.getenv("NO_ML") == "true", reason="No numpy in NO_ML")
def test_config_wild_values():
    import numpy as np
    with CliRunner().isolated_filesystem():
        conf = config.Config(run_dir=".")
        conf['numpy'] = np.random.normal(size=(10,))
        conf['class'] = CliRunner()
        conf['list'] = [np.random.normal(1), 32.5]
        conf['dir'] = {"numpy": np.random.normal(size=(10,))}

        conf2 = config.Config(config_paths=['config.yaml'])
        assert len(conf2['numpy']) == 10
        assert len(conf2['dir']['numpy']) == 10
        assert "CliRunner" in conf2['class']
        assert conf2['list'][-1] == 32.5

def test_config_persists():
    with CliRunner().isolated_filesystem():
        conf = config.Config(run_dir='.')
        conf['a'] = 5
        conf.b = 14.3

        conf2 = config.Config(config_paths=['config.yaml'])
        assert conf2['a'] == 5
        assert conf2.b == 14.3


def test_config_defaults():
    with CliRunner().isolated_filesystem():
        open('config-defaults.yaml', 'w').write(textwrap.dedent("""\
            wandb_version: 1

            a:
              desc: the number nineteen
              value: 19
            b:
              desc: null
              value: a_cow
            c: 141912
            """))
        conf = config.Config(wandb_dir='.')
        assert conf.a == 19
        assert conf.b == 'a_cow'
        assert conf.c == 141912


def test_config_file_overrides():
    with CliRunner().isolated_filesystem():
        open('config-defaults.yaml', 'w').write(textwrap.dedent("""\
            wandb_version: 1

            a:
              desc: the number nineteen
              value: 19
            b:
              desc: null
              value: a_cow
            """))
        open('config-special.yaml', 'w').write(textwrap.dedent("""\
            wandb_version: 1

            a:
              desc: the number nineteen
              value: 43
            """))
        conf = config.Config(wandb_dir='.', config_paths=[
                             'config-special.yaml'])
        assert conf.a == 43
        assert conf.b == 'a_cow'



import wandb
from .utils import git_repo
from gym import core
from gym.wrappers.monitoring.video_recorder import ImageEncoder
import pytest
import sys


@pytest.mark.skipif(sys.version_info < (3, 0), reason="gym no longer supports python 2.7")
def test_patch(wandb_init_run, git_repo):
    wandb.gym.monitor()
    with open("test.gif", "w") as f:
        f.write("test")
    ir = ImageEncoder("test.gif", (28, 28, 3), 10, 10)
    ir.close()
    assert wandb_init_run.summary["videos"]['_type'] == "video-file"



import collections
import os
import threading

import wandb
from wandb import util

EventJobDone = collections.namedtuple('EventJobDone', ('job', 'success'))


class UploadJob(threading.Thread):
    def __init__(self, done_queue, stats, api, save_name, path, artifact_id, md5, copied, save_fn, digest):
        """A file upload thread.

        Arguments:
            done_queue: queue.Queue in which to put an EventJobDone event when
                the upload finishes.
            push_function: function(save_name, actual_path) which actually uploads
                the file.
            save_name: string logical location of the file relative to the run
                directory.
            path: actual string path of the file to upload on the filesystem.
        """
        self._done_queue = done_queue
        self._stats = stats
        self._api = api
        self.save_name = save_name
        self.save_path = self.path = path
        self.artifact_id = artifact_id
        self.md5 = md5
        self.copied = copied
        self.save_fn = save_fn
        self.digest = digest
        super(UploadJob, self).__init__()

    def run(self):
        success = False
        try:
            success = self.push()
        finally:
            if self.copied and os.path.isfile(self.save_path):
                os.remove(self.save_path)
            self._done_queue.put(EventJobDone(self, success))

    def push(self):
        try:
            size = os.path.getsize(self.save_path)
        except OSError:
            size = 0

        if self.save_fn:
            # Retry logic must happen in save_fn currently
            try:
                deduped = self.save_fn(
                    lambda _, t: self._stats.update_uploaded_file(self.save_path, t))
            except Exception as e:
                self._stats.update_failed_file(self.save_path)
                wandb.util.sentry_exc(e)
                message = str(e)
                # TODO: this is usually XML, but could be JSON
                if hasattr(e, "response"):
                    message = e.response.content
                wandb.termerror('Error uploading "{}": {}, {}'.format(
                    self.save_path, type(e).__name__, message))
                return False

            if deduped:
                self._stats.set_file_deduped(self.save_path)
            return True

        if self.md5:
            # This is the new artifact manifest upload flow, in which we create the
            # database entry for the manifest file before creating it. This is used for
            # artifact L0 files. Which now is only artifact_manifest.json
            response = self._api.create_artifact_manifest(self.save_name, self.md5, self.artifact_id)
            upload_url = response['uploadUrl']
            upload_headers = response['uploadHeaders']
        else:
            # The classic file upload flow. We get a signed url and upload the file
            # then the backend handles the cloud storage metadata callback to create the
            # file entry. This flow has aged like a fine wine.
            project = self._api.get_project()
            _, upload_headers, result = self._api.upload_urls(project, [self.save_name])
            file_info = result[self.save_name]
            upload_url = file_info['url']

        if upload_url is None:
            self._stats.set_file_deduped(self.save_name)
        else:
            extra_headers = {}
            for upload_header in upload_headers:
                key, val = upload_header.split(':', 1)
                extra_headers[key] = val
            # Copied from push TODO(artifacts): clean up
            # If the upload URL is relative, fill it in with the base URL,
            # since its a proxied file store like the on-prem VM.
            if upload_url.startswith('/'):
                upload_url = '{}{}'.format(self._api.api_url, upload_url)
            try:
                with open(self.save_path, 'rb') as f:
                    self._api.upload_file_retry(
                        upload_url,
                        f,
                        lambda _, t: self.progress(t),
                        extra_headers=extra_headers)
            except Exception as e:
                self._stats.update_failed_file(self.save_name)
                wandb.util.sentry_exc(e)
                wandb.termerror('Error uploading "{}": {}, {}'.format(
                    self.save_name, type(e).__name__, e))
                return False
        return True

    def progress(self, total_bytes):
        self._stats.update_uploaded_file(self.save_name, total_bytes)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_wandb
----------------------------------

Tests for the `wandb.apis.PublicApi` module.
"""
import datetime
import pytest
import os
import yaml
import tempfile
from .api_mocks import *
from click.testing import CliRunner
import git
import json
from .utils import git_repo, runner
import h5py
import numpy as np

import wandb
from wandb import Api
from six import StringIO
api = Api()


def test_parse_path_simple():
    u, p, r = api._parse_path("user/proj/run")
    assert u == "user"
    assert p == "proj"
    assert r == "run"


def test_parse_path_leading():
    u, p, r = api._parse_path("/user/proj/run")
    assert u == "user"
    assert p == "proj"
    assert r == "run"


def test_parse_path_docker():
    u, p, r = api._parse_path("user/proj:run")
    assert u == "user"
    assert p == "proj"
    assert r == "run"


def test_parse_path_docker_proj(mock_server):
    u, p, r = api._parse_path("proj:run")
    assert u == "vanpelt"
    assert p == "proj"
    assert r == "run"


def test_parse_path_url():
    u, p, r = api._parse_path("user/proj/runs/run")
    assert u == "user"
    assert p == "proj"
    assert r == "run"


def test_parse_path_user_proj(mock_server):
    u, p, r = api._parse_path("proj/run")
    assert u == "vanpelt"
    assert p == "proj"
    assert r == "run"


def test_parse_path_proj(mock_server):
    u, p, r = api._parse_path("proj")
    assert u == "vanpelt"
    assert p == "proj"
    assert r == "proj"


def test_run_from_path(mock_server):
    api.flush()
    run = api.run("test/test/test")
    assert run.summary_metrics == {"acc": 100, "loss": 0}


def test_run_retry(mock_server):
    api.flush()
    mock_server.set_context("fail_times", 2)
    run = api.run("test/test/test")
    assert run.summary_metrics == {"acc": 100, "loss": 0}


def test_run_history(mock_server):
    api.flush()
    run = api.run("test/test/test")
    assert run.history(pandas=False)[0] == {'acc': 10, 'loss': 90}


def test_run_config(mock_server):
    api.flush()
    run = api.run("test/test/test")
    assert run.config == {'epochs': 10}


def test_run_history_system(mock_server):
    api.flush()
    run = api.run("test/test/test")
    assert run.history(stream="system", pandas=False) == [
        {'cpu': 10}, {'cpu': 20}, {'cpu': 30}]


def test_run_summary(mock_server):
    api.flush()
    run = api.run("test/test/test")
    run.summary.update({"cool": 1000})
    assert json.loads(mock_server.ctx["graphql"][-1]["variables"]["summaryMetrics"]) == {"acc": 100, "loss": 0, "cool": 1000}


def test_run_create(mock_server):
    run = api.create_run(project="test")
    assert mock_server.ctx["graphql"][-1]["variables"] == {'entity': 'vanpelt', 'name': run.id, 'project': 'test'}


def test_run_update(mock_server):
    api.flush()
    run = api.run("test/test/test")
    run.tags.append("test")
    run.config["foo"] = "bar"
    run.update()
    assert json.loads(mock_server.ctx["graphql"][-1]["variables"]["summaryMetrics"]) == {"acc": 100, "loss": 0}
    assert mock_server.ctx["graphql"][-2]["variables"]["entity"] == "test"


def test_run_files(runner, mock_server):
    with runner.isolated_filesystem():
        api.flush()
        run = api.run("test/test/test")
        file = run.files()[0]
        file.download()
        assert os.path.exists("weights.h5")
        raised = False
        try:
            file.download()
        except wandb.CommError:
            raised = True
        assert raised


def test_run_file(runner, mock_server):
    with runner.isolated_filesystem():
        api.flush()
        run = api.run("test/test/test")
        file = run.file("weights.h5")
        assert not os.path.exists("weights.h5")
        print("YO", file.url)
        file.download()
        assert os.path.exists("weights.h5")


def test_runs_from_path(mock_server):
    api.flush()
    runs = api.runs("test/test")
    assert len(runs) == 4
    list(runs)
    assert len(runs.objects) == 2
    assert runs[0].summary_metrics == {"acc": 100, "loss": 0}


def test_runs_from_path_index(mock_server):
    api.flush()
    mock_server.set_context("page_times", 4)
    runs = api.runs("test/test")
    assert len(runs) == 4
    print(list(runs))
    assert runs[3]
    assert len(runs.objects) == 4


def test_projects(mock_server):
    projects = api.projects("test")
    # projects doesn't provide a length for now, so we iterate
    # them all to count
    count = 0
    for proj in projects:
        count += 1
    assert count == 2


def test_sweep(request_mocker):
    run_responses = [random_run_response() for _ in range(5)]
    sweep_response = random_sweep_response()
    sweep_response['runs'] = {'edges': [{'node': r} for r in run_responses]}
    for r in run_responses:
        r['sweepName'] = sweep_response['name']

    mock_graphql_request(request_mocker, {'project': {'sweep': sweep_response}})

    sweep = api.sweep('test/test/{}'.format(sweep_response['name']))
    assert sweep.entity == 'test'
    assert sweep.project == 'test'
    assert sweep.name
    assert sweep.config
    assert sweep.best_loss
    assert len(sweep.runs) == len(run_responses)


def test_reports(request_mocker):
    report_response = basic_report_response()
    mock_graphql_request(request_mocker, {'project': report_response}, body_match='query Run')
    report = api.reports("test/test/test")[0]
    assert report.sections[0]['name'] == '01: Effect of hidden layer size on feedforward net'
    query = {"op": "OR",
             "filters": [{"op": "AND",
                          "filters": [{"key": {"section": "tags", "name": "basic_feedforward"},
                                       "op": "=",
                                       "value": True,
                                       "disabled": False}]}]}
    print(report.query_generator.filter_to_mongo(query))
    assert report.query_generator.filter_to_mongo(query) == {'$or': [{'$and': [{'tags': 'basic_feedforward'}]}]}


def test_run_sweep(request_mocker):
    run_response = random_run_response()
    sweep_response = random_sweep_response()
    run_response['sweepName'] = sweep_response['name']

    mock_graphql_request(request_mocker, {'project': {'run': run_response}}, body_match='query Run')
    mock_graphql_request(request_mocker, {'project': {'sweep': sweep_response}}, body_match='query Sweep')

    run = api.run('test/test/{}'.format(run_response['name']))
    assert run.id
    assert run.name
    assert run.sweep
    assert run.sweep.id
    assert run.sweep.runs == [run]


def test_runs_sweeps(request_mocker):
    """Request a bunch of runs from different sweeps at the same time.
    Ensure each run's sweep attribute is set to the appropriate value.
    """
    api = Api()
    run_responses = [random_run_response() for _ in range(7)]
    sweep_a_response = random_sweep_response()
    sweep_b_response = random_sweep_response()
    run_responses[0]['sweepName'] = sweep_b_response['name']
    run_responses[1]['sweepName'] = sweep_a_response['name']
    run_responses[3]['sweepName'] = sweep_b_response['name']
    run_responses[4]['sweepName'] = sweep_a_response['name']
    run_responses[5]['sweepName'] = sweep_b_response['name']
    run_responses[6]['sweepName'] = sweep_b_response['name']

    runs_response = {
        'project': {
            'runCount': len(run_responses),
            'runs': {
                'pageInfo': {'hasNextPage': False},
                'edges': [{'node': r, 'cursor': 'cursor'} for r in run_responses],
            }
        }
    }
    mock_graphql_request(request_mocker, runs_response, body_match='query Runs')
    mock_graphql_request(request_mocker, {'project': {'sweep': sweep_a_response}},
                         body_match=sweep_a_response['name'])
    mock_graphql_request(request_mocker, {'project': {'sweep': sweep_b_response}},
                         body_match=sweep_b_response['name'])

    runs = list(api.runs('test/test'))
    sweep_a = runs[1].sweep
    sweep_b = runs[0].sweep
    assert len(runs) == len(run_responses)
    assert sweep_a.runs == [runs[1], runs[4]]
    assert sweep_b.runs == [runs[0], runs[3], runs[5], runs[6]]
    assert runs[0].sweep is sweep_b
    assert runs[1].sweep is sweep_a
    assert runs[2].sweep is None
    assert runs[3].sweep is sweep_b
    assert runs[4].sweep is sweep_a
    assert runs[5].sweep is sweep_b
    assert runs[6].sweep is sweep_b


# @pytest.mark.skip(readon='fails when I run the whole suite, but not when I run just this file')
def test_read_advanced_summary(runner, request_mocker, upsert_run, query_download_h5, query_upload_h5):
    with runner.isolated_filesystem():
        run = run_response()
        run["summaryMetrics"] = json.dumps({
            "special": {"_type": "numpy.ndarray", "min": 0, "max": 20},
            "normal": 32,
            "nested": {"deep": {"_type": "numpy.ndarray", "min": 0, "max": 20}}})
        query_mocker('project', {'run': run})(request_mocker)
        file = os.path.join(tempfile.gettempdir(), "test.h5")
        with h5py.File(file, 'w') as h5:
            h5["summary/special"] = np.random.rand(100)
            h5["summary/nested.deep"] = np.random.rand(100)
        query_download_h5(request_mocker, content=open(file, "rb").read())
        api.flush()
        run = api.run("test/test/test")
        assert len(run.summary["special"]) == 100
        assert len(run.summary["nested"]["deep"]) == 100
        update_mock = upsert_run(request_mocker)
        h5_mock = query_upload_h5(request_mocker)
        run.summary.update({"nd_time": np.random.rand(1000)})
        assert len(run.summary["nd_time"]) == 1000
        # TODO: this passes locally, but fails consistently in CI?!?
        #assert h5_mock.called
        del run.summary["nd_time"]
        run.summary.open_h5()
        assert list(sorted(run.summary._h5["summary"].keys())) == [
            "nested.deep", "special"]

def test_artifact_versions(runner, mock_server):
    versions = api.artifact_versions("dataset", "mnist")
    assert len(versions) == 2
    assert versions[0].name == "mnist:v0"
    assert versions[1].name == "mnist:v1"

def test_artifact_type(runner, mock_server):
    atype = api.artifact_type("dataset")
    assert atype.name == "dataset"
    col = atype.collection("mnist")
    assert col.name == "mnist"
    cols = atype.collections()
    assert cols[0].name == "mnist"

def test_artifact_types(runner, mock_server):
    atypes = api.artifact_types("dataset")

    raised = False
    try:
        assert len(atypes) == 2
    except ValueError:
        raised = True
    assert raised
    assert atypes[0].name == "dataset"

def test_artifact_get_path(runner, mock_server):
    art = api.artifact("entity/project/mnist:v0", type="dataset")
    assert art.type == "dataset"
    assert art.name == "mnist:v0"
    with runner.isolated_filesystem():
        path = art.get_path("digits.h5")
        res = path.download()
        assert res == os.path.expanduser("~")+ "/.cache/wandb/artifacts/obj/md5/4d/e489e31c57834a21b8be7111dab613"

def test_artifact_file(runner, mock_server):
    with runner.isolated_filesystem():
        art = api.artifact("entity/project/mnist:v0", type="dataset")
        path = art.file()
        assert path == "./artifacts/mnist:v0/digits.h5"

def test_artifact_download(runner, mock_server):
    with runner.isolated_filesystem():
        art = api.artifact("entity/project/mnist:v0", type="dataset")
        path = art.download()
        assert path == "./artifacts/mnist:v0"

def test_artifact_run_used(runner, mock_server):
    api.flush()
    run = api.run("test/test/test")
    arts = run.used_artifacts()
    assert len(arts) == 2
    assert arts[0].name == "abc123"

def test_artifact_run_logged(runner, mock_server):
    api.flush()
    run = api.run("test/test/test")
    arts = run.logged_artifacts()
    assert len(arts) == 2
    assert arts[0].name == "abc123"


import os
import pytest

from click.testing import CliRunner
import wandb.util as util
from wandb.settings import Settings
from wandb import env


def test_read_empty_settings():
    settings = Settings()
    assert settings.get(Settings.DEFAULT_SECTION, 'foo', fallback=None) is None


def test_read_global_setting(global_wandb_settings):
    global_wandb_settings.write("[default]\nfoo = bar\n")
    global_wandb_settings.flush()

    settings = Settings()
    assert settings.get(Settings.DEFAULT_SECTION, 'foo') == 'bar'


def test_read_local_setting(global_wandb_settings, local_wandb_settings):
    global_wandb_settings.write("[default]\nfoo = baz\n")
    global_wandb_settings.flush()

    local_wandb_settings.write("[default]\nfoo = bar\n")
    local_wandb_settings.flush()

    settings = Settings()
    assert settings.get(Settings.DEFAULT_SECTION, 'foo') == 'bar'


def test_write_setting_globally(global_wandb_settings):
    settings = Settings()
    settings.set(Settings.DEFAULT_SECTION, 'foo', 'bar', globally=True, persist=True)

    with open(global_wandb_settings.name, "r") as f:
        data = f.read()
        assert "[default]" in data
        assert "foo = bar" in data


def test_write_setting_locally(local_wandb_settings):
    settings = Settings()
    settings.set(Settings.DEFAULT_SECTION, 'foo', 'bar', persist=True)

    with open(local_wandb_settings.name, "r") as f:
        data = f.read()
        assert "[default]" in data
        assert "foo = bar" in data


def test_items(global_wandb_settings, local_wandb_settings):
    global_wandb_settings.write("[default]\nfoo = baz\n")
    global_wandb_settings.flush()

    local_wandb_settings.write("[default]\nfoo = bar\n")
    local_wandb_settings.flush()

    settings = Settings()

    assert settings.items() == {
        'section': Settings.DEFAULT_SECTION,
        'foo': 'bar',
    }


@pytest.fixture
def global_wandb_settings(tmpdir):
    os.environ[env.CONFIG_DIR] = tmpdir.strpath

    with open(os.path.join(tmpdir.strpath, 'settings'), "w+") as f:
        yield f

    del os.environ[env.CONFIG_DIR]


@pytest.fixture
def local_wandb_settings():
    with CliRunner().isolated_filesystem():
        util.mkdir_exists_ok('wandb')
        with open(os.path.join('wandb', 'settings'), 'w+') as f:
            yield f



from copy import deepcopy

import tensorflow as tf

from wandb import util
from wandb.apis.file_stream import Chunk
from wandb.data_types import history_dict_to_json
from wandb.tensorboard import *
import wandb
try:
    from tensorflow.train import summary_iterator
except ImportError:
    from tensorflow.compat.v1.train import summary_iterator


if hasattr(tf.estimator, 'SessionRunHook'):
    # In tf 1.14 and beyond, SessionRunHook is in the estimator package.
    SessionRunHook = tf.estimator.SessionRunHook
else:
    # In older versions it's in train.
    SessionRunHook = tf.train.SessionRunHook


class WandbHook(SessionRunHook):
    def __init__(self, summary_op=None, steps_per_log=1000, history=None):
        self._summary_op = summary_op
        self._steps_per_log = steps_per_log
        # TODO(adrian): might be better to set this to wandb.run.history here
        # because that is the de facto default.
        self._history = history

    def begin(self):
        if self._summary_op is None:
            self._summary_op = tf.summary.merge_all()
        self._step = -1

    def before_run(self, run_context):
        self._step += 1
        return tf.train.SessionRunArgs({"summary": self._summary_op})

    def after_run(self, run_context, run_values):
        if self._step % self._steps_per_log == 0:
            log(run_values.results["summary"], history=self._history)


def stream_tfevents(path, file_api, run, step=0, namespace=""):
    """Parses and streams a tfevents file to the server"""
    last_step = 0
    row = {}
    buffer = []
    last_row = {}
    global_step_key = namespaced_tag("global_step", namespace)
    try:
        for summary in summary_iterator(path):
            parsed = tf_summary_to_dict(summary, namespace=namespace)
            if parsed is None:
                continue
            if last_step != parsed[global_step_key]:
                last_step = parsed[global_step_key]
                if len(row) > 3:  # Must have more than _timestamp, _step, and global_step
                    step += 1
                    row["_step"] = step
                    last_row = history_dict_to_json(run, deepcopy(row))
                    file_api.push("wandb-history.jsonl", util.json_dumps_safer_history(last_row))
                    row = {}
            row.update(parsed)
    except tf.errors.DataLossError:
        wandb.termwarn("Found a truncated record in tfevents file, stopping parse")
    step += 1
    row["_step"] = step
    last_row = history_dict_to_json(run, deepcopy(row))
    file_api.push("wandb-history.jsonl", util.json_dumps_safer_history(last_row))
    return last_row


__all__ = ['log', 'patch', 'stream_tfevents', 'WandbHook']



# -*- coding: utf-8 -*-
"""Utility routines
"""

def sweepwarn(s):
    print("[WARNING]", s)


def sweeplog(s):
    print("[LOG]", s)
        

def sweeperror(s):
    print("[ERROR]", s)


def sweepdebug(s):
    print("[DEBUG]", s)


def pad(lst, num, val=None, left=False):
    pad = [val] * (num - len(lst))
    lst = pad + lst if left else lst + pad
    return lst




import base64
import os
import shutil

from wandb import env
from wandb import util

class ArtifactsCache(object):
    def __init__(self, cache_dir):
        self._cache_dir = cache_dir
        util.mkdir_exists_ok(self._cache_dir)
        self._md5_obj_dir = os.path.join(self._cache_dir, 'obj', 'md5')
        self._etag_obj_dir = os.path.join(self._cache_dir, 'obj', 'etag')

    def check_md5_obj_path(self, b64_md5, size):
        hex_md5 = util.bytes_to_hex(base64.b64decode(b64_md5))
        path = os.path.join(self._cache_dir, 'obj', 'md5', hex_md5[:2], hex_md5[2:])
        if os.path.isfile(path) and os.path.getsize(path) == size:
            return path, True
        util.mkdir_exists_ok(os.path.dirname(path))
        return path, False

    def check_etag_obj_path(self, etag, size):
        path = os.path.join(self._cache_dir, 'obj', 'etag', etag[:2], etag[2:])
        if os.path.isfile(path) and os.path.getsize(path) == size:
            return path, True
        util.mkdir_exists_ok(os.path.dirname(path))
        return path, False

_artifacts_cache = None

def get_artifacts_cache():
    global _artifacts_cache
    if _artifacts_cache is None:
        cache_dir = os.path.join(env.get_cache_dir(), 'artifacts')
        _artifacts_cache = ArtifactsCache(cache_dir)
    return _artifacts_cache



from collections import OrderedDict

# six >=1.13.0 not available in all environmnents yet
#from six.moves.collections_abc import Sequence
try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence

import inspect
import logging
import os
import sys
import six
import types
import yaml
import platform

import wandb
from wandb import env

FNAME = 'config.yaml'

logger = logging.getLogger(__name__)


class ConfigError(wandb.Error):
    pass


def boolify(s):
    if s.lower() == 'none':
        return None
    if s.lower() == 'true':
        return True
    if s.lower() == 'false':
        return False
    raise ValueError("Not a boolean")

def is_kaggle():
    return os.getenv("KAGGLE_KERNEL_RUN_TYPE") != None or "kaggle_environments" in sys.modules

def huggingface_version():
    if "transformers" in sys.modules:
        trans = wandb.util.get_module("transformers")
        if hasattr(trans, "__version__"):
            return trans.__version__
    return None

class Config(object):
    """
    Config is a dict like object that corresponds to the hyperparameters for a training run.

    Config values for a run will be saved to app.wandb.ai or the local wandb server and
    displayed in the runs table along with summary values.

    Config values should only be set once.

    Config values can be set like a dictionary config['a'] = 1 or set with config.a = 1.

    Nested config values are supported.

    Example:
        ```
        wandb.init()
        config = wandb.config
        config.learning_rate = 0.5
        config.batch_size = 100

        train_model(config)
        ```
    """

    def __init__(self, config_paths=[], wandb_dir=None, run_dir=None):
        object.__setattr__(self, '_wandb_dir', wandb_dir)

        # OrderedDict to make writing unit tests easier. (predictable order for
        # .key())
        object.__setattr__(self, '_items', OrderedDict())
        object.__setattr__(self, '_descriptions', {})

        self._load_defaults()
        self._load_wandb()
        for conf_path in config_paths:
            self._load_file(conf_path)
        self._set_wandb('cli_version', wandb.__version__)
        self._set_wandb('python_version', platform.python_version())
        self._set_wandb('is_jupyter_run', wandb._get_python_type() != "python")
        self._set_wandb('is_kaggle_kernel', is_kaggle())
        hf_version = huggingface_version()
        if hf_version:
            self._set_wandb('huggingface_version', hf_version)

        # Do this after defaults because it triggers loading of pre-existing
        # config.yaml (if it exists)
        self.set_run_dir(run_dir)

        self.persist()

    def _telemetry_update(self):
        """Add telemetry data to internal config structure."""
        updated = False

        # detect framework by checking what is loaded
        loaded = {}
        loaded['lightgbm'] = sys.modules.get('lightgbm')
        loaded['catboost'] = sys.modules.get('catboost')
        loaded['xgboost'] = sys.modules.get('xgboost')
        loaded['fastai'] = sys.modules.get('fastai')
        loaded['torch'] = sys.modules.get('torch')
        loaded['keras'] = sys.modules.get('keras')  # vanilla keras
        loaded['tensorflow'] = sys.modules.get('tensorflow')
        loaded['sklearn'] = sys.modules.get('sklearn')
        # TODO(jhr): tfkeras is always loaded with recent tensorflow
        #loaded['tfkeras'] = sys.modules.get('tensorflow.python.keras')

        priority = ('lightgbm', 'catboost', 'xgboost', 'fastai', 'torch', 'keras', 'tfkeras', 'tensorflow', 'sklearn')
        framework = next((f for f in priority if loaded.get(f)), None)
        if framework:
            self._set_wandb('framework', framework)
            updated = True

        return updated

    @classmethod
    def from_environment_or_defaults(cls):
        conf_paths = os.environ.get(env.CONFIG_PATHS, [])
        run_dir = os.environ.get(env.RUN_DIR)
        if conf_paths:
            conf_paths = conf_paths.split(',')
        return Config(config_paths=conf_paths, wandb_dir=wandb.wandb_dir(), run_dir=run_dir)

    def _load_wandb(self):
        # We load docker into the config from the env
        image = os.getenv(env.DOCKER)
        digest = None
        if image:
            if "@" in image:
                image, digest = image.split("@")
            self._set_wandb("docker_image", image)
            if digest:
                self._set_wandb("docker_digest", digest)

    def _set_wandb(self, key, val):
        self._items["_wandb"] = self._items.get("_wandb", {})
        self._items["_wandb"][key] = val

    def _load_defaults(self):
        if not self._wandb_dir:
            logger.debug('wandb dir not provided, skipping defaults')
            return
        defaults_path = os.path.join('config-defaults.yaml')
        if not os.path.exists(defaults_path):
            logger.debug('no defaults not found in %s' % defaults_path)
            return
        self._load_file(defaults_path)

    def _load_file(self, path):
        subkey = None
        if '::' in path:
            conf_path, subkey = path.split('::', 1)
        else:
            conf_path = path
        try:
            conf_file = open(conf_path)
        except (OSError, IOError):
            raise ConfigError('Couldn\'t read config file: %s' % conf_path)
        try:
            loaded = wandb.util.load_yaml(conf_file)
        except yaml.parser.ParserError:
            raise ConfigError('Invalid YAML in config-defaults.yaml')
        if subkey:
            try:
                loaded = loaded[subkey]
            except KeyError:
                raise ConfigError('Asked for {} but {} not present in {}'.format(
                    path, subkey, conf_path))
        for key, val in loaded.items():
            if key == 'wandb_version':
                continue
            if isinstance(val, dict):
                if 'value' not in val:
                    raise ConfigError('In config {} value of {} is dict, but does not contain "value" key'.format(
                        path, key))
                self._items[key] = val['value']
                if 'desc' in val:
                    self._descriptions[key] = val['desc']
            else:
                self._items[key] = val

    def _load_values(self):
        """Load config.yaml from the run directory if available."""
        path = self._config_path()
        if path is not None and os.path.isfile(path):
            self._load_file(path)

    def _config_path(self):
        if self._run_dir and os.path.isdir(self._run_dir):
            return os.path.join(self._run_dir, FNAME)
        return None

    def keys(self):
        """All keys in the current configuration"""
        return [k for k in self._items.keys() if k != '_wandb']

    def desc(self, key):
        """The description of a given key"""
        return self._descriptions.get(key)

    def load_json(self, json):
        """Loads existing config from JSON"""
        for key in json:
            if key == "wandb_version" or key == "_wandb":
                continue
            self._items[key] = json[key].get('value')
            self._descriptions[key] = json[key].get('desc')

    def set_run_dir(self, run_dir):
        """Set the run directory to which this Config should be persisted.

        Changes to this Config won't be written anywhere unless the run directory
        is set.
        """
        object.__setattr__(self, '_run_dir', run_dir)
        self._load_values()

    def persist(self):
        """Stores the current configuration for pushing to W&B"""
        # In dryrun mode, without wandb run, we don't
        # save config  on initial load, because the run directory
        # may not be created yet (because we don't know if we're
        # being used in a run context, or as an API).
        # TODO: Defer saving somehow, maybe via an events system
        path = self._config_path()
        if path is None:
            return
        with open(path, "w") as conf_file:
            conf_file.write(str(self))
        if wandb.run and wandb.run._jupyter_agent:
            wandb.run._jupyter_agent.start()

    def get(self, *args):
        return self._items.get(*args)

    def __getitem__(self, key):
        return self._items[key]

    def __setitem__(self, key, val):
        key, val = self._sanitize(key, val)
        self._items[key] = val
        self.persist()

    __setattr__ = __setitem__

    def __getattr__(self, key):
        return self.__getitem__(key)

    def _sanitize(self, key, val, allow_val_change=False):
        # We always normalize keys by stripping '-'
        key = key.strip('-')
        val = self._sanitize_val(val)
        if not allow_val_change:
            if key in self._items and val != self._items[key]:
                raise ConfigError('Attempted to change value of key "{}" from {} to {}\nIf you really want to do this, pass allow_val_change=True to config.update()'.format(
                    key, self._items[key], val))
        return key, val

    def _sanitize_val(self, val):
        """Turn all non-builtin values into something safe for YAML"""
        if isinstance(val, dict):
            converted = {}
            for key, value in six.iteritems(val):
                converted[key] = self._sanitize_val(value)
            return converted
        if isinstance(val, slice):
            converted = dict(slice_start=val.start, slice_step=val.step, slice_stop=val.stop)
            return converted
        val, _ = wandb.util.json_friendly(val)
        if isinstance(val, Sequence) and not isinstance(val, six.string_types):
            converted = []
            for value in val:
                converted.append(self._sanitize_val(value))
            return converted
        else:
            if val.__class__.__module__ not in ('builtins', '__builtin__'):
                val = str(val)
            return val

    def _update(self, params, allow_val_change=False, as_defaults=False, exclude_keys=None, include_keys=None):
        exclude_keys = exclude_keys or []
        include_keys = include_keys or []
        params = params or {}
        if not isinstance(params, dict):
            # Handle some cases where params is not a dictionary
            # by trying to convert it into a dictionary
            meta = inspect.getmodule(params)
            if meta:
                is_tf_flags_module = isinstance(
                    params, types.ModuleType) and meta.__name__ == 'tensorflow.python.platform.flags'
                if is_tf_flags_module or meta.__name__ == 'absl.flags':
                    params = params.FLAGS
                    meta = inspect.getmodule(params)

            # newer tensorflow flags (post 1.4) uses absl.flags
            if meta and meta.__name__ == "absl.flags._flagvalues":
                params = {name: params[name].value for name in dir(params)}
            elif "__flags" in vars(params):
                # for older tensorflow flags (pre 1.4)
                if not '__parsed' in vars(params):
                    params._parse_flags()
                params = vars(params)['__flags']
            elif not hasattr(params, '__dict__'):
                raise TypeError(
                    "config must be a dict or have a __dict__ attribute.")
            else:
                # params is a Namespace object (argparse)
                # or something else
                params = vars(params)

        if not isinstance(params, dict):
            raise ConfigError('Expected dict but received %s' % params)
        if exclude_keys and include_keys:
            raise ConfigError('Expected at most only one of exclude_keys or include_keys')
        for key, val in params.items():
            if key in exclude_keys:
                continue
            if include_keys and key not in include_keys:
                continue
            key, val = self._sanitize(key, val, allow_val_change=allow_val_change or as_defaults)
            if as_defaults and key in self._items:
                continue
            self._items[key] = val
        self.persist()

    def update(self, params, allow_val_change=False, exclude_keys=None, include_keys=None):
        self._update(params,
                exclude_keys=exclude_keys,
                include_keys=include_keys,
                allow_val_change=allow_val_change)

    def setdefaults(self, params, exclude_keys=None, include_keys=None):
        self._update(params,
                exclude_keys=exclude_keys,
                include_keys=include_keys,
                as_defaults=True)
        return dict(self)

    def setdefault(self, key, default=None):
        key, val = self._sanitize(key, default, allow_val_change=True)
        if key in self._items:
            return self._items[key]
        self._items[key] = val
        self.persist()
        return val

    def as_dict(self):
        defaults = {}
        for key, val in self._items.items():
            defaults[key] = {'value': val,
                             'desc': self._descriptions.get(key)}
        return defaults

    def user_items(self):
        """Retrieve user configured config parameters as a key value tuple generator"""
        for key, val in self._items.items():
            if key != '_wandb':
                yield (key, val)

    def __str__(self):
        s = b"wandb_version: 1"
        as_dict = self.as_dict()
        if as_dict:  # adding an empty dictionary here causes a parse error
            s += b'\n\n' + yaml.dump(as_dict, Dumper=yaml.SafeDumper, default_flow_style=False,
                                     allow_unicode=True, encoding='utf-8')
        return s.decode("utf-8")


class ConfigStatic(object):
    def __init__(self, config):
        object.__setattr__(self, "__dict__", dict(config))

    def __setattr__(self, name, value):
        raise AttributeError("Error: wandb.run.config_static is a readonly object")

    def __setitem__(self, key, val):
        raise AttributeError("Error: wandb.run.config_static is a readonly object")

    def keys(self):
        return self.__dict__.keys()

    def __getitem__(self, key):
        return self.__dict__[key]

    def __str__(self):
        return str(self.__dict__)



import wandb
from wandb import util
from wandb.plots.utils import test_missing, test_types, encode_labels
chart_limit = wandb.Table.MAX_ROWS

def heatmap(x_labels, y_labels, matrix_values, show_text=False):
        """
        Generates a heatmap.

        Arguments:
         matrix_values (arr): 2D dataset of shape x_labels * y_labels, containing
                            heatmap values that can be coerced into an ndarray.
         x_labels  (list): Named labels for rows (x_axis).
         y_labels  (list): Named labels for columns (y_axis).
         show_text (bool): Show text values in heatmap cells.

        Returns:
         Nothing. To see plots, go to your W&B run page then expand the 'media' tab
               under 'auto visualizations'.

        Example:
         wandb.log({'heatmap': wandb.plots.HeatMap(x_labels, y_labels,
                    matrix_values)})
        """
        np = util.get_module("numpy", required="roc requires the numpy library, install with `pip install numpy`")
        scikit = util.get_module("sklearn", required="roc requires the scikit library, install with `pip install scikit-learn`")

        if (test_missing(x_labels=x_labels, y_labels=y_labels,
            matrix_values=matrix_values) and test_types(x_labels=x_labels,
            y_labels=y_labels, matrix_values=matrix_values)):
            matrix_values = np.array(matrix_values)
            wandb.termlog('Visualizing heatmap.')

            def heatmap_table(x_labels, y_labels, matrix_values, show_text):
                x_axis=[]
                y_axis=[]
                values=[]
                count = 0
                for i, x in enumerate(x_labels):
                    for j, y in enumerate(y_labels):
                        x_axis.append(x)
                        y_axis.append(y)
                        values.append(matrix_values[j][i])
                        count+=1
                        if count >= chart_limit:
                            wandb.termwarn("wandb uses only the first %d datapoints to create the plots."% wandb.Table.MAX_ROWS)
                            break
                if show_text:
                    heatmap_key = 'wandb/heatmap/v1'
                else:
                    heatmap_key = 'wandb/heatmap_no_text/v1'
                return wandb.visualize(
                    heatmap_key, wandb.Table(
                    columns=['x_axis', 'y_axis', 'values'],
                    data=[
                        [x_axis[i], y_axis[i], round(values[i], 2)] for i in range(len(x_axis))
                    ]
                ))
            return heatmap_table(x_labels, y_labels, matrix_values, show_text)



import collections
import json
import base64
import hashlib
import re
import os
import tempfile
import time
import shutil
import requests
from six.moves.urllib.parse import urlparse

from wandb.compat import tempfile as compat_tempfile

from wandb.apis import InternalApi, Progress, CommError
from wandb import artifacts_cache
from wandb import util
from wandb import file_pusher
from wandb.core import termwarn, termlog


def md5_string(string):
    hash_md5 = hashlib.md5()
    hash_md5.update(string.encode())
    return base64.b64encode(hash_md5.digest()).decode('ascii')


def md5_hash_file(path):
    hash_md5 = hashlib.md5()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(64 * 1024), b""):
            hash_md5.update(chunk)
    return hash_md5


def md5_file_b64(path):
    return base64.b64encode(md5_hash_file(path).digest()).decode('ascii')


def md5_file_hex(path):
    return md5_hash_file(path).hexdigest()


class ServerManifestV1(object):
    """This implements the same artifact digest algorithm as the server."""

    # hash must be the base64 encoded md5 checksum of the file at path.
    ArtifactManifestEntry = collections.namedtuple('ArtifactManifestEntry', ('path', 'hash'))

    def __init__(self, entries):
        # lexicographic sort. paths must be unique, so sort doesn't ever visit hash
        self.entries = sorted(entries)


class Artifact(object):
    """An artifact object you can write files into, and pass to log_artifact."""

    # A local manifest contains the path to the local file in addition to the path within
    # the artifact.
    LocalArtifactManifestEntry = collections.namedtuple('LocalArtifactManifestEntry', (
        'path', 'hash', 'local_path'))

    def __init__(self, name, type, description=None, metadata=None):
        if not re.match('^[a-zA-Z0-9_\-.]+$', name):
            raise ValueError('Artifact name may only contain alphanumeric characters, dashes, underscores, and dots. Invalid name: "%s"' % name)
        if type is None:
            raise ValueError("type is required when logging artifacts, specify \"dataset\", \"model\", or a custom type")
        # TODO: this shouldn't be a property of the artifact. It's a more like an
        # argument to log_artifact.
        self._storage_policy = WandbStoragePolicy()
        self._file_specs = {}
        self._api = InternalApi() # TODO: persist project in settings?
        self._final = False
        self._digest = None
        self._file_entries = None
        self._manifest = ArtifactManifestV1(self, self._storage_policy)
        self._cache = artifacts_cache.get_artifacts_cache()
        self._added_new = False
        # You can write into this directory when creating artifact files
        self._artifact_dir = compat_tempfile.TemporaryDirectory(missing_ok_on_cleanup=True)
        self.server_manifest = None
        self.type = type
        self.name = name
        self.description = description
        self.metadata = metadata

    @property
    def id(self):
        # The artifact hasn't been saved so an ID doesn't exist yet.
        return None

    @property
    def entity(self):
        # TODO: querying for default entity a good idea here?
        return self._api.settings('entity') or self._api.viewer().get("entity")

    @property
    def project(self):
        return self._api.settings('project')

    @property
    def manifest(self):
        self.finalize()
        return self._manifest

    # TODO: Currently this returns the L0 digest. Is this what we want?
    @property
    def digest(self):
        self.finalize()
        return self._digest

    def _ensure_can_add(self):
        if self._final:
            raise ValueError('Can\'t add to finalized artifact.')

    def new_file(self, name):
        self._ensure_can_add()
        path = os.path.join(self._artifact_dir.name, name.lstrip('/'))
        if os.path.exists(path):
            raise ValueError('File with name "%s" already exists' % name)
        util.mkdir_exists_ok(os.path.dirname(path))
        self._added_new = True
        return open(path, 'w')

    def add_file(self, local_path, name=None):
        self._ensure_can_add()
        if not os.path.isfile(local_path):
            raise ValueError('Path is not a file: %s' % local_path)

        name = name or os.path.basename(local_path)
        entry = ArtifactManifestEntry(
            name, None, digest=md5_file_b64(local_path),
            size=os.path.getsize(local_path),
            local_path=local_path)
        self._manifest.add_entry(entry)

    def add_dir(self, local_path, name=None):
        self._ensure_can_add()
        if not os.path.isdir(local_path):
            raise ValueError('Path is not a directory: %s' % local_path)

        termlog('Adding directory to artifact (%s)... ' %
            os.path.join('.', os.path.normpath(local_path)), newline=False)
        start_time = time.time()

        paths = []
        for dirpath, _, filenames in os.walk(local_path, followlinks=True):
            for fname in filenames:
                physical_path = os.path.join(dirpath, fname)
                logical_path = os.path.relpath(physical_path, start=local_path)
                if name is not None:
                    logical_path = os.path.join(name, logical_path)
                paths.append((logical_path, physical_path))

        def add_manifest_file(log_phy_path):
            logical_path, physical_path = log_phy_path
            self._manifest.add_entry(
                ArtifactManifestEntry(
                    logical_path,
                    None,
                    digest=md5_file_b64(physical_path),
                    size=os.path.getsize(physical_path),
                    local_path=physical_path
                )
            )

        import multiprocessing.dummy  # this uses threads
        NUM_THREADS = 8
        pool = multiprocessing.dummy.Pool(NUM_THREADS)
        pool.map(add_manifest_file, paths)
        pool.close()
        pool.join()

        termlog('Done. %.1fs' % (time.time() - start_time), prefix=False)

    def add_reference(self, uri, name=None, checksum=True, max_objects=None):
        url = urlparse(uri)
        if not url.scheme:
            raise ValueError('References must be URIs. To reference a local file, use file://')
        if self._final:
            raise ValueError('Can\'t add to finalized artifact.')
        manifest_entries = self._storage_policy.store_reference(
            self, uri, name=name, checksum=checksum, max_objects=max_objects)
        for entry in manifest_entries:
            self._manifest.add_entry(entry)

    def get_path(self, name):
        raise ValueError('Cannot load paths from an artifact before it has been saved')

    def download(self):
        raise ValueError('Cannot call download on an artifact before it has been saved')

    def finalize(self):
        if self._final:
            return self._file_entries

        # Record any created files in the manifest.
        if self._added_new:
            self.add_dir(self._artifact_dir.name)

        # mark final after all files are added
        self._final = True

        # Add the manifest itself as a file.
        with tempfile.NamedTemporaryFile('w+', suffix=".json", delete=False) as fp:
            json.dump(self._manifest.to_manifest_json(), fp, indent=4)
            self._file_specs['wandb_manifest.json'] = fp.name
            manifest_file = fp.name

        # Calculate the server manifest
        file_entries = []
        for name, local_path in self._file_specs.items():
            file_entries.append(self.LocalArtifactManifestEntry(
                name, md5_file_b64(local_path), os.path.abspath(local_path)))
        self.server_manifest = ServerManifestV1(file_entries)
        self._digest = self._manifest.digest()

        # If there are new files, move them into the artifact cache now. Our temp
        # self._artifact_dir may not be available by the time file pusher syncs
        # these files.
        if self._added_new:
            # Update the file entries for new files to point at their new location.
            def remap_entry(entry):
                if entry.local_path is None or not entry.local_path.startswith(self._artifact_dir.name):
                    return entry
                rel_path = os.path.relpath(entry.local_path, start=self._artifact_dir.name)
                local_path = os.path.join(self._artifact_dir.name, rel_path)
                cache_path, hit = self._cache.check_md5_obj_path(entry.digest, entry.size)
                if not hit:
                    shutil.copyfile(local_path, cache_path)
                entry.local_path = cache_path
            for entry in self._manifest.entries.values():
                remap_entry(entry)

class ArtifactSaver(object):
    def __init__(self, api, digest, server_manifest_entries, manifest_json, file_pusher=None, is_user_created=False):
        # NOTE: manifest_entries are LocalManifestEntry but they get converted to
        # arrays when we convert to json, so we need to access fields by index instead
        # of by name
        self._api = api
        self._file_pusher = file_pusher
        self._digest = digest
        self._server_manifest_entries = server_manifest_entries
        self._manifest = ArtifactManifest.from_manifest_json(None, manifest_json)
        self._is_user_created = is_user_created
        self._server_artifact = None

    def save(self, type, name, metadata=None, description=None, aliases=None, labels=None, use_after_commit=False):
        aliases = aliases or []
        alias_specs = []
        for alias in aliases:
            if ":" in alias:
                # Users can explicitly alias this artifact to names
                # other than the primary one passed in by using the
                # 'secondaryName:alias' notation.
                idx = alias.index(":")
                artifact_collection_name = alias[:idx-1]
                tag = alias[idx+1:]
            else:
                artifact_collection_name = name
                tag = alias
            alias_specs.append({
                'artifactCollectionName': artifact_collection_name,
                'alias': tag,
            })

        """Returns the server artifact."""
        self._server_artifact = self._api.create_artifact(
            type,
            name,
            self._digest,
            metadata=metadata,
            aliases=alias_specs, labels=labels, description=description,
            is_user_created=self._is_user_created)
        # TODO(artifacts):
        #   if it's committed, all is good. If it's committing, just moving ahead isn't necessarily
        #   correct. It may be better to poll until it's committed or failed, and then decided what to
        #   do
        artifact_id = self._server_artifact['id']
        if self._server_artifact['state'] == 'COMMITTED' or self._server_artifact['state'] == 'COMMITTING':
            # TODO: update aliases, labels, description etc?
            if use_after_commit:
                self._api.use_artifact(artifact_id)
            return self._server_artifact
        elif self._server_artifact['state'] != 'PENDING':
            # TODO: what to do in this case?
            raise Exception('Server artifact not in PENDING state', self._server_artifact)

        # Upload Artifact "L0" files. This should only be artifact_manifest.json. We need to use
        # the use_prepare_flow, so that the file entry is created in our database before the
        # upload to cloud storage commences
        for path, hash, local_path in self._server_manifest_entries:
            # We need to use the "use_prepare_flow" option
            self._file_pusher.file_changed(path, local_path, artifact_id, use_prepare_flow=True)

        step_prepare = file_pusher.step_prepare.StepPrepare(self._api, 0.1, 0.01, 1000)  # TODO: params
        step_prepare.start()

        # Upload Artifact "L1" files, the actual artifact contents
        self._file_pusher.store_manifest_files(
            self._manifest,
            artifact_id,
            lambda entry, progress_callback : self._manifest.storage_policy.store_file(
                artifact_id,
                entry,
                step_prepare,
                progress_callback=progress_callback))

        def on_commit():
            if use_after_commit:
                self._api.use_artifact(artifact_id)
            step_prepare.shutdown()

        # This will queue the commit. It will only happen after all the file uploads are done
        self._file_pusher.commit_artifact(artifact_id, on_commit=on_commit)
        return self._server_artifact

class ArtifactManifest(object):

    @classmethod
    # TODO: we don't need artifact here.
    def from_manifest_json(cls, artifact, manifest_json):
        if 'version' not in manifest_json:
            raise ValueError('Invalid manifest format. Must contain version field.')
        version = manifest_json['version']
        for sub in cls.__subclasses__():
            if sub.version() == version:
                return sub.from_manifest_json(artifact, manifest_json)

    @classmethod
    def version(cls):
        pass

    def __init__(self, artifact, storage_policy, entries=None):
        self.artifact = artifact
        self.storage_policy = storage_policy
        self.entries = entries or {}

    def to_manifest_json(self):
        raise NotImplementedError()

    def digest(self):
        raise NotImplementedError()

    def add_entry(self, entry):
        if entry.path in self.entries:
            raise ValueError('Cannot add the same path twice: %s' % entry.path)
        self.entries[entry.path] = entry


class ArtifactManifestV1(ArtifactManifest):

    @classmethod
    def version(cls):
        return 1

    @classmethod
    def from_manifest_json(cls, artifact, manifest_json):
        if manifest_json['version'] != cls.version():
            raise ValueError('Expected manifest version 1, got %s' % manifest_json['version'])

        storage_policy_name = manifest_json['storagePolicy']
        storage_policy_config = manifest_json.get('storagePolicyConfig', {})
        storage_policy_cls = StoragePolicy.lookup_by_name(storage_policy_name)
        if storage_policy_cls is None:
            raise ValueError('Failed to find storage policy "%s"' % storage_policy_name)

        entries = {
            name: ArtifactManifestEntry(
                name, val.get('ref'), val['digest'],
                size=val.get('size'),
                extra=val.get('extra'),
                local_path=val.get('local_path'))
            for name, val in manifest_json['contents'].items()
        }

        return cls(artifact, storage_policy_cls.from_config(storage_policy_config), entries)

    def __init__(self, artifact, storage_policy, entries=None):
        super(ArtifactManifestV1, self).__init__(artifact, storage_policy, entries=entries)

    def to_manifest_json(self, include_local=False):
        """This is the JSON that's stored in wandb_manifest.json
        
        If include_local is True we also include the local paths to files. This is
        used to represent an artifact that's waiting to be saved on the current
        system. We don't need to include the local paths in the artifact manifest
        contents.
        """
        contents = {}
        for entry in sorted(self.entries.values(), key=lambda k: k.path):
            json_entry = {
                'digest': entry.digest,
            }
            if entry.ref is not None:
                json_entry['ref'] = entry.ref
            if entry.extra:
                json_entry['extra'] = entry.extra
            if entry.size is not None:
                json_entry['size'] = entry.size
            if include_local and entry.local_path is not None:
                json_entry['local_path'] = entry.local_path
            contents[entry.path] = json_entry
        return {
            'version': self.__class__.version(),
            'storagePolicy': self.storage_policy.name(),
            'storagePolicyConfig': self.storage_policy.config() or {},
            'contents': contents
        }

    def digest(self):
        hasher = hashlib.md5()
        hasher.update("wandb-artifact-manifest-v1\n".encode())
        for (name, entry) in sorted(self.entries.items(), key=lambda kv: kv[0]):
            hasher.update('{}:{}\n'.format(name, entry.digest).encode())
        return hasher.hexdigest()


class ArtifactManifestEntry(object):

    def __init__(self, path, ref, digest, size=None, extra=None, local_path=None):
        self.path = path
        self.ref = ref  # This is None for files stored in the artifact.
        self.digest = digest
        self.size = size
        self.extra = extra or {}
        # This is not stored in the manifest json, it's only used in the process
        # of saving
        self.local_path = local_path

    def __repr__(self):
        if self.ref is not None:
            summary = 'ref: %s/%s' % (self.ref, self.path)
        else:
            summary = 'digest: %s' % self.digest

        return "<ManifestEntry %s>" % summary


class StoragePolicy(object):

    @classmethod
    def lookup_by_name(cls, name):
        for sub in cls.__subclasses__():
            if sub.name() == name:
                return sub
        return None

    @classmethod
    def name(cls):
        pass

    @classmethod
    def from_config(cls, config):
        pass

    def config(self):
        pass

    def load_file(self, artifact, name, manifest_entry):
        raise NotImplementedError

    def store_file(self, artifact_id, entry, preparer, progress_callback=None):
        raise NotImplementedError

    def store_reference(self, artifact, path, name=None, checksum=True, max_objects=None):
        raise NotImplementedError

    def load_reference(self, artifact, name, manifest_entry, local=False):
        raise NotImplementedError


class WandbStoragePolicy(StoragePolicy):

    @classmethod
    def name(cls):
        return 'wandb-storage-policy-v1'

    @classmethod
    def from_config(cls, config):
        return cls()

    def __init__(self):
        s3 = S3Handler()
        gcs = GCSHandler()
        file_handler = LocalFileHandler()

        self._api = InternalApi()
        self._handler = MultiHandler(handlers=[
            s3,
            gcs,
            file_handler,
        ], default_handler=TrackingHandler())

        self._cache = artifacts_cache.get_artifacts_cache()

        # I believe this makes the first sleep 1s, and then doubles it up to
        # total times, which makes for ~18 hours.
        retry_strategy = requests.packages.urllib3.util.retry.Retry(
            backoff_factor=1,
            total=16,
            status_forcelist=(308, 409, 429, 500, 502, 503, 504))
        self._session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(
            max_retries=retry_strategy,
            pool_connections=64,
            pool_maxsize=64)
        self._session.mount('http://', adapter)
        self._session.mount('https://', adapter)

    def config(self):
        return None

    def load_file(self, artifact, name, manifest_entry):
        path, hit = self._cache.check_md5_obj_path(manifest_entry.digest, manifest_entry.size)
        if hit:
            return path

        response = self._session.get(
            self._file_url(
                self._api,
                artifact.entity,
                manifest_entry.digest),
            auth=("api", self._api.api_key),
            stream=True)
        response.raise_for_status()

        with open(path, "wb") as file:
            for data in response.iter_content(chunk_size=16 * 1024):
                file.write(data)
        return path

    def store_reference(self, artifact, path, name=None, checksum=True, max_objects=None):
        return self._handler.store_path(artifact, path, name=name, checksum=checksum, max_objects=max_objects)

    def load_reference(self, artifact, name, manifest_entry, local=False):
        return self._handler.load_path(self._cache, manifest_entry, local)

    def _file_url(self, api, entity_name, md5):
        md5_hex = util.bytes_to_hex(base64.b64decode(md5))
        return '{}/artifacts/{}/{}'.format(api.settings("base_url"), entity_name, md5_hex)

    def store_file(self, artifact_id, entry, preparer, progress_callback=None):
        # write-through cache
        cache_path, hit = self._cache.check_md5_obj_path(entry.digest, entry.size)
        if not hit:
            shutil.copyfile(entry.local_path, cache_path)

        resp = preparer.prepare(lambda: {
            "artifactID": artifact_id,
            "name": entry.path,
            "md5": entry.digest,
        })

        exists = resp.upload_url is None
        if not exists:
            with open(entry.local_path, "rb") as file:
                # This fails if we don't send the first byte before the signed URL
                # expires.
                r = self._session.put(resp.upload_url,
                                 headers={
                                     header.split(":", 1)[0]: header.split(":", 1)[1]
                                     for header in (resp.upload_headers or {})
                                 },
                                 data=Progress(file, callback=progress_callback))
                r.raise_for_status()
        return exists


# Don't use this yet!
class __S3BucketPolicy(StoragePolicy):

    @classmethod
    def name(cls):
        return 'wandb-s3-bucket-policy-v1'

    @classmethod
    def from_config(cls, config):
        if 'bucket' not in config:
            raise ValueError('Bucket name not found in config')
        return cls(config['bucket'])

    def __init__(self, bucket):
        self._bucket = bucket
        s3 = S3Handler(bucket)
        local = LocalFileHandler()

        self._handler = MultiHandler(handlers=[
            s3,
            local,
        ], default_handler=TrackingHandler())

    def config(self):
        return {
            'bucket': self._bucket
        }

    def load_path(self, artifact, manifest_entry, local=False):
        return self._handler.load_path(artifact, manifest_entry, local=local)

    def store_path(self, artifact, path, name=None, checksum=True, max_objects=None):
        return self._handler.store_path(artifact, path, name=name, checksum=checksum, max_objects=max_objects)


class StorageHandler(object):

    def scheme(self):
        """
        :return: The scheme to which this handler applies.
        :rtype: str
        """
        pass

    def load_path(self, artifact, manifest_entry, local=False):
        """
        Loads the file or directory within the specified artifact given its
        corresponding index entry.

        :param manifest_entry: The index entry to load
        :type manifest_entry: ArtifactManifestEntry
        :return: A path to the file represented by `index_entry`
        :rtype: os.PathLike
        """
        pass

    def store_path(self, artifact, path, name=None, checksum=True, max_objects=None):
        """
        Stores the file or directory at the given path within the specified artifact.

        :param path: The path to store
        :type path: str
        :param name: If specified, the logical name that should map to `path`
        :type name: str
        :return: A list of manifest entries to store within the artifact
        :rtype: list(ArtifactManifestEntry)
        """
        pass


class MultiHandler(StorageHandler):

    def __init__(self, handlers=None, default_handler=None):
        self._handlers = {}
        self._default_handler = default_handler

        handlers = handlers or []
        for handler in handlers:
            self._handlers[handler.scheme] = handler

    @property
    def scheme(self):
        raise NotImplementedError()

    def load_path(self, artifact, manifest_entry, local=False):
        url = urlparse(manifest_entry.ref)
        if url.scheme not in self._handlers:
            if self._default_handler is not None:
                return self._default_handler.load_path(artifact, manifest_entry, local=local)
            raise ValueError('No storage handler registered for scheme "%s"' % url.scheme)
        return self._handlers[url.scheme].load_path(artifact, manifest_entry, local=local)

    def store_path(self, artifact, path, name=None, checksum=True, max_objects=None):
        url = urlparse(path)
        if url.scheme not in self._handlers:
            if self._handlers is not None:
                return self._default_handler.store_path(artifact, path, name=name, checksum=checksum, max_objects=max_objects)
            raise ValueError('No storage handler registered for scheme "%s"' % url.scheme)
        return self._handlers[url.scheme].store_path(artifact, path, name=name, checksum=checksum, max_objects=max_objects)


class TrackingHandler(StorageHandler):

    def __init__(self, scheme=None):
        """
        Tracks paths as is, with no modification or special processing. Useful
        when paths being tracked are on file systems mounted at a standardized
        location.

        For example, if the data to track is located on an NFS share mounted on
        /data, then it is sufficient to just track the paths.
        """
        self._scheme = scheme

    @property
    def scheme(self):
        return self._scheme

    def load_path(self, artifact, manifest_entry, local=False):
        if local:
            # Likely a user error. The tracking handler is
            # oblivious to the underlying paths, so it has
            # no way of actually loading it.
            url = urlparse(manifest_entry.ref)
            raise ValueError('Cannot download file at path %s, scheme %s not recognized' %
                             (manifest_entry.ref, url.scheme))
        return manifest_entry.path

    def store_path(self, artifact, path, name=None, checksum=False, max_objects=None):
        url = urlparse(path)
        if name is None:
            raise ValueError('You must pass name="<entry_name>" when tracking references with unknown schemes. ref: %s' % path)
        termwarn('Artifact references with unsupported schemes cannot be checksummed: %s' % path)
        name = name or url.path[1:]  # strip leading slash
        return [ArtifactManifestEntry(name, path, digest=path)]

DEFAULT_MAX_OBJECTS = 10000

class LocalFileHandler(StorageHandler):

    """Handles file:// references"""

    def __init__(self, scheme=None):
        """
        Tracks files or directories on a local filesystem. Directories
        are expanded to create an entry for each file contained within.
        """
        self._scheme = scheme or "file"
        self._cache = artifacts_cache.get_artifacts_cache()

    @property
    def scheme(self):
        return self._scheme

    def load_path(self, artifact, manifest_entry, local=False):
        url = urlparse(manifest_entry.ref)
        local_path = '%s%s' % (url.netloc, url.path)
        if not os.path.exists(local_path):
            raise ValueError('Local file reference: Failed to find file at path %s.' % local_path)

        path, hit = self._cache.check_md5_obj_path(
            manifest_entry.digest, manifest_entry.size)
        if hit:
            return path
        md5 = md5_file_b64(local_path)
        if md5 != manifest_entry.digest:
            raise ValueError('Local file reference: Digest mismatch for path %s. expected %s but found %s' %
                             (local_path, manifest_entry.digest, md5))

        util.mkdir_exists_ok(os.path.dirname(path))
        shutil.copy(local_path, path)
        return path

    def store_path(self, artifact, path, name=None, checksum=True, max_objects=None):
        url = urlparse(path)
        local_path = '%s%s' % (url.netloc, url.path)
        max_objects = max_objects or DEFAULT_MAX_OBJECTS
        # We have a single file or directory
        # Note, we follow symlinks for files contained within the directory
        entries = []
        if checksum == False:
            return [ArtifactManifestEntry(name or os.path.basename(path), path, digest=path)]

        if os.path.isdir(local_path):
            i = 0
            start_time = time.time()
            termlog('Generating checksum for up to %i files in "%s"...' % (max_objects, local_path), newline=False)
            for root, dirs, files in os.walk(local_path):
                for sub_path in files:
                    i += 1
                    if i >= max_objects:
                        raise ValueError('Exceeded %i objects tracked, pass max_objects to add_reference' % max_objects)
                    entry = ArtifactManifestEntry(os.path.basename(sub_path),
                        os.path.join(path, sub_path), size=os.path.getsize(sub_path), digest=md5_file_b64(sub_path))
                    entries.append(entry)
            termlog('Done. %.1fs' % (time.time() - start_time), prefix=False)
        elif os.path.isfile(local_path):
            name = name or os.path.basename(local_path)
            entry = ArtifactManifestEntry(name, path, size=os.path.getsize(local_path), digest=md5_file_b64(local_path))
            entries.append(entry)
        else:
            # TODO: update error message if we don't allow directories.
            raise ValueError('Path "%s" must be a valid file or directory path' % path)
        return entries

class S3Handler(StorageHandler):

    def __init__(self, scheme=None):
        self._scheme = scheme or "s3"
        self._s3 = None
        self._versioning_enabled = None
        self._cache = artifacts_cache.get_artifacts_cache()

    @property
    def scheme(self):
        return self._scheme

    def init_boto(self):
        if self._s3 is not None:
            return self._s3
        boto3 = util.get_module('boto3', required="s3:// references requires the boto3 library, run pip install wandb[aws]")
        self._s3 = boto3.session.Session().resource('s3', endpoint_url=os.getenv("AWS_S3_ENDPOINT_URL"), region_name=os.getenv("AWS_REGION"))
        self._botocore = util.get_module("botocore")
        return self._s3

    def _parse_uri(self, uri):
        url = urlparse(uri)
        bucket = url.netloc
        key = url.path[1:] # strip leading slash
        return bucket, key

    def versioning_enabled(self, bucket):
        self.init_boto()
        if self._versioning_enabled is not None:
            return self._versioning_enabled
        res = self._s3.BucketVersioning(bucket)
        self._versioning_enabled = res.status == 'Enabled'
        return self._versioning_enabled

    def load_path(self, artifact, manifest_entry, local=False):
        path, hit = self._cache.check_etag_obj_path(manifest_entry.digest, manifest_entry.size)
        if hit:
            return path

        self.init_boto()
        bucket, key = self._parse_uri(manifest_entry.ref)
        version = manifest_entry.extra.get('versionID')

        extra_args = {}
        if version is None:
            # We don't have version information so just get the latest version
            # and fallback to listing all versions if we don't have a match.
            obj = self._s3.Object(bucket, key)
            etag = self._etag_from_obj(obj)
            if etag != manifest_entry.digest:
                if self.versioning_enabled(bucket):
                    # Fallback to listing versions
                    obj = None
                    object_versions = self._s3.Bucket(bucket).object_versions.filter(Prefix=key)
                    for object_version in object_versions:
                        if manifest_entry.extra.get('etag') == object_version.e_tag[1:-1]:
                            obj = object_version.Object()
                            extra_args['VersionId'] = object_version.version_id
                            break
                    if obj is None:
                        raise ValueError("Couldn't find object version for %s/%s matching etag %s" %
                            (self._bucket, key, manifest_entry.extra.get('etag')))
                else:
                    raise ValueError('Digest mismatch for object %s: expected %s but found %s' %
                                    (manifest_entry.ref, manifest_entry.digest, etag))
        else:
            obj = self._s3.ObjectVersion(bucket, key, version).Object()
            extra_args['VersionId'] = version

        if not local:
            return manifest_entry.ref

        obj.download_file(path, ExtraArgs=extra_args)
        return path

    def store_path(self, artifact, path, name=None, checksum=True, max_objects=None):
        self.init_boto()
        bucket, key = self._parse_uri(path)
        max_objects = max_objects or DEFAULT_MAX_OBJECTS
        if checksum == False:
            return [ArtifactManifestEntry(name or key, path, digest=path)]

        objs = [self._s3.Object(bucket, key)]
        start_time = None
        multi = False
        try:
            objs[0].load()
        except self._botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                multi = True
                start_time = time.time()
                termlog('Generating checksum for up to %i objects with prefix "%s"... ' % (max_objects, key), newline=False)
                objs = self._s3.Bucket(bucket).objects.filter(Prefix=key).limit(max_objects)
            else:
                raise CommError("Unable to connect to S3 (%s): %s" % (e.response['Error']['Code'], e.response['Error']['Message']))

        # Weird iterator scoping makes us assign this to a local function
        size = self._size_from_obj
        entries = [self._entry_from_obj(obj, path, name, prefix=key, multi=multi) for obj in objs if size(obj) > 0]
        if start_time is not None:
            termlog('Done. %.1fs' % (time.time() - start_time), prefix=False)
        if len(entries) >= max_objects:
            raise ValueError('Exceeded %i objects tracked, pass max_objects to add_reference' % max_objects)
        return entries

    def _size_from_obj(self, obj):
        # ObjectSummary has size, Object has content_length
        if hasattr(obj, "size"):
            size = obj.size
        else:
            size = obj.content_length
        return size

    def _entry_from_obj(self, obj, path, name=None, prefix="", multi=False):
        ref = path
        if name is None:
            if prefix in obj.key and prefix != obj.key:
                name = os.path.relpath(obj.key, start=prefix)
                ref = os.path.join(path, name)
            else:
                name = os.path.basename(obj.key)
        elif multi:
            # We're listing a path and user provided name, just prepend it
            name = os.path.join(name, os.path.basename(obj.key))
            ref = os.path.join(path, name)
        return ArtifactManifestEntry(name, ref,
            self._etag_from_obj(obj), size=self._size_from_obj(obj), extra=self._extra_from_obj(obj))

    @staticmethod
    def _etag_from_obj(obj):
        return obj.e_tag[1:-1]  # escape leading and trailing quote

    @staticmethod
    def _extra_from_obj(obj):
        extra = {
            'etag': obj.e_tag[1:-1],  # escape leading and trailing quote
        }
        # ObjectSummary will never have version_id
        if hasattr(obj, "version_id") and obj.version_id != "null":
            extra['versionID'] = obj.version_id
        return extra

    @staticmethod
    def _content_addressed_path(md5):
        # TODO: is this the structure we want? not at all human
        # readable, but that's probably OK. don't want people
        # poking around in the bucket
        return 'wandb/%s' % base64.b64encode(md5.encode("ascii")).decode("ascii")

class GCSHandler(StorageHandler):
    def __init__(self, scheme=None):
        self._scheme = scheme or "gs"
        self._client = None
        self._versioning_enabled = None
        self._cache = artifacts_cache.get_artifacts_cache()

    def versioning_enabled(self, bucket):
        if self._versioning_enabled is not None:
            return self._versioning_enabled
        self.init_gcs()
        bucket = self._client.bucket(bucket)
        bucket.reload()
        self._versioning_enabled = bucket.versioning_enabled
        return self._versioning_enabled

    @property
    def scheme(self):
        return self._scheme

    def init_gcs(self):
        if self._client is not None:
            return self._client
        storage = util.get_module('google.cloud.storage', required="gs:// references requires the google-cloud-storage library, run pip install wandb[gcp]")
        self._client = storage.Client()
        return self._client

    def _parse_uri(self, uri):
        url = urlparse(uri)
        bucket = url.netloc
        key = url.path[1:]
        return bucket, key

    def load_path(self, artifact, manifest_entry, local=False):
        path, hit = self._cache.check_md5_obj_path(
            manifest_entry.digest, manifest_entry.size)
        if hit:
            return path

        self.init_gcs()
        bucket, key = self._parse_uri(manifest_entry.ref)
        version = manifest_entry.extra.get('versionID')

        extra_args = {}
        obj = None
        # First attempt to get the generation specified, this will return None if versioning is not enabled
        if version is not None:
            obj = self._client.bucket(bucket).get_blob(key, generation=version)

        if obj is None:
            # Object versioning is disabled on the bucket, so just get
            # the latest version and make sure the MD5 matches.
            obj = self._client.bucket(bucket).get_blob(key)
            if obj is None:
                raise ValueError('Unable to download object %s with generation %s' % (manifest_entry.ref, version))
            md5 = obj.md5_hash
            if md5 != manifest_entry.digest:
                raise ValueError('Digest mismatch for object %s: expected %s but found %s' %
                    (manifest_entry.ref, manifest_entry.digest, md5))

        if not local:
            return manifest_entry.ref

        obj.download_to_filename(path)
        return path

    def store_path(self, artifact, path, name=None, checksum=True, max_objects=None):
        self.init_gcs()
        bucket, key = self._parse_uri(path)
        max_objects = max_objects or DEFAULT_MAX_OBJECTS

        if checksum == False:
            return [ArtifactManifestEntry(name or key, path, digest=path)]
        start_time = None
        obj = self._client.bucket(bucket).get_blob(key)
        multi = obj is None
        if multi:
            start_time = time.time()
            termlog('Generating checksum for up to %i objects with prefix "%s"... ' % (max_objects, key), newline=False)
            objects = self._client.bucket(bucket).list_blobs(prefix=key, max_results=max_objects)
        else:
            objects = [obj]

        entries = [self._entry_from_obj(obj, path, name, prefix=key, multi=multi) for obj in objects]
        if start_time is not None:
            termlog('Done. %.1fs' % (time.time() - start_time), prefix=False)
        if len(entries) >= max_objects:
            raise ValueError('Exceeded %i objects tracked, pass max_objects to add_reference' % max_objects)
        return entries

    def _entry_from_obj(self, obj, path, name=None, prefix="", multi=False):
        ref = path
        if name is None:
            if prefix in obj.name and prefix != obj.name:
                name = os.path.relpath(obj.name, start=prefix)
                ref = os.path.join(path, name)
            else:
                name = os.path.basename(obj.name)
        elif multi:
            # We're listing a path and user provided name, just prepend it
            name = os.path.join(name, os.path.basename(obj.name))
            ref = os.path.join(path, name)
        return ArtifactManifestEntry(name, ref, obj.md5_hash, size=obj.size, extra=self._extra_from_obj(obj))

    @staticmethod
    def _extra_from_obj(obj):
        return {
            'etag': obj.etag,
            'versionID': obj.generation,
        }

    @staticmethod
    def _content_addressed_path(md5):
        # TODO: is this the structure we want? not at all human
        # readable, but that's probably OK. don't want people
        # poking around in the bucket
        return 'wandb/%s' % base64.b64encode(md5.encode("ascii")).decode("ascii")



"""Batching file prepare requests to our API."""

import multiprocessing
import multiprocessing.pool
import collections
import os
import shutil
import threading
import time
from six.moves import queue
import wandb.util

from wandb.filesync import step_upload


RequestUpload = collections.namedtuple(
    'RequestUpload', ('path', 'save_name', 'artifact_id', 'copy', 'use_prepare_flow', 'save_fn', 'digest'))
RequestStoreManifestFiles = collections.namedtuple(
    'RequestStoreManifestFiles', ('manifest', 'artifact_id', 'save_fn'))
RequestCommitArtifact = collections.namedtuple(
    'RequestCommitArtifact', ('artifact_id', 'on_commit'))
RequestFinish = collections.namedtuple('RequestFinish', ())

    
class StepChecksum(object):
    def __init__(self, api, tempdir, request_queue, output_queue, stats):
        self._api = api
        self._tempdir = tempdir
        self._request_queue = request_queue
        self._output_queue = output_queue
        self._stats = stats

        self._thread = threading.Thread(target=self._thread_body)
        self._thread.daemon = True

    def _thread_body(self):
        finished = False
        while True:
            req = self._request_queue.get()
            if isinstance(req, RequestUpload):
                path = req.path
                if req.copy:
                    path = os.path.join(self._tempdir.name, '%s-%s' % (
                        wandb.util.generate_id(), req.save_name))
                    wandb.util.mkdir_exists_ok(os.path.dirname(path))
                    shutil.copy2(req.path, path)
                checksum = None
                if req.use_prepare_flow:
                    # passing a checksum through indicates that we'd like to use the
                    # "prepare" file upload flow, in which we prepare the files in
                    # the database before uploading them. This is currently only
                    # used for artifact manifests
                    checksum = wandb.util.md5_file(path)
                self._stats.init_file(req.save_name, os.path.getsize(path))
                self._output_queue.put(
                    step_upload.RequestUpload(
                        path, req.save_name, req.artifact_id, checksum, req.copy,
                        req.save_fn, req.digest))
            elif isinstance(req, RequestStoreManifestFiles):
                for entry in req.manifest.entries.values():
                    if entry.local_path:
                        # This stupid thing is needed so the closure works correctly.
                        def make_save_fn_with_entry(save_fn, entry):
                            return lambda progress_callback: save_fn(entry, progress_callback)
                        self._stats.init_file(entry.local_path, entry.size, is_artifact_file=True)
                        self._output_queue.put(
                            step_upload.RequestUpload(
                                entry.local_path,
                                entry.path,
                                req.artifact_id,
                                entry.digest,
                                False,
                                make_save_fn_with_entry(req.save_fn, entry),
                                entry.digest))
            elif isinstance(req, RequestCommitArtifact):
                self._output_queue.put(step_upload.RequestCommitArtifact(req.artifact_id, req.on_commit))
            elif isinstance(req, RequestFinish):
                break
            else:
                raise Exception('internal error')

        self._output_queue.put(step_upload.RequestFinish())

    def start(self):
        self._thread.start()

    def is_alive(self):
        return self._thread.is_alive()

    def finish(self):
        self._request_queue.put(RequestFinish())



from wandb import util, Image, Error, termwarn

# assums X represents images and y_true/y_pred are logits for each class
def image_categorizer_dataframe(x, y_true, y_pred, labels, example_ids=None):
    np = util.get_module('numpy', required='dataframes require numpy')
    pd = util.get_module('pandas', required='dataframes require pandas')

    x, y_true, y_pred, labels = np.array(x), np.array(y_true), np.array(y_pred), np.array(labels)

    # If there is only one output value of true_prob, convert to 2 class false_prob, true_prob
    if y_true[0].shape[-1] == 1 and y_pred[0].shape[-1] == 1:
        y_true = np.concatenate((1-y_true, y_true), axis=-1)
        y_pred = np.concatenate((1-y_pred, y_pred), axis=-1)

    if x.shape[0] != y_true.shape[0]:
        termwarn('Sample count mismatch: x(%d) != y_true(%d). skipping evaluation' % (x.shape[0], y_true.shape[0]))
        return
    if x.shape[0] != y_pred.shape[0]:
        termwarn('Sample count mismatch: x(%d) != y_pred(%d). skipping evaluation' % (x.shape[0], y_pred.shape[0]))
        return
    if y_true.shape[-1] != len(labels):
        termwarn('Label count mismatch: y_true(%d) != labels(%d). skipping evaluation' % (y_true.shape[-1], len(labels)))
        return
    if y_pred.shape[-1] != len(labels):
        termwarn('Label count mismatch: y_pred(%d) != labels(%d). skipping evaluation' % (y_pred.shape[-1], len(labels)))
        return

    class_preds = []
    for i in range(len(labels)):
        class_preds.append(y_pred[:,i])

    images = [Image(img) for img in x]
    true_class = labels[y_true.argmax(axis=-1)]
    true_prob = y_pred[np.arange(y_pred.shape[0]), y_true.argmax(axis=-1)]
    pred_class = labels[y_pred.argmax(axis=-1)]
    pred_prob = y_pred[np.arange(y_pred.shape[0]), y_pred.argmax(axis=-1)]
    correct = true_class == pred_class

    if example_ids is None:
        example_ids = ['example_' + str(i) for i in range(len(x))]

    dfMap = {
        'wandb_example_id': example_ids,
        'image': images,
        'true_class': true_class,
        'true_prob': true_prob,
        'pred_class': pred_class,
        'pred_prob': pred_prob,
        'correct': correct,
    }

    for i in range(len(labels)):
        dfMap['prob_{}'.format(labels[i])] = class_preds[i]

    all_columns = [
        'wandb_example_id',
        'image',
        'true_class',
        'true_prob',
        'pred_class',
        'pred_prob',
        'correct',
    ] + ['prob_{}'.format(l) for l in labels]

    return pd.DataFrame(dfMap, columns=all_columns)

def image_segmentation_dataframe(x, y_true, y_pred, labels=None, example_ids=None, class_colors=None):
    np = util.get_module('numpy', required='dataframes require numpy')
    y_pred = np.array(y_pred)
    if y_pred[0].shape[-1] == 1:
        return image_segmentation_binary_dataframe(x, y_true, y_pred, example_ids=example_ids)
    else:
        return image_segmentation_multiclass_dataframe(x, y_true, y_pred, labels=labels, example_ids=example_ids, class_colors=class_colors)

def image_segmentation_binary_dataframe(x, y_true, y_pred, example_ids=None):
    np = util.get_module('numpy', required='dataframes require numpy')
    pd = util.get_module('pandas', required='dataframes require pandas')

    x, y_true, y_pred= np.array(x), np.array(y_true), np.array(y_pred)

    if x.shape[0] != y_true.shape[0]:
        termwarn('Sample count mismatch: x(%d) != y_true(%d). skipping evaluation' % (x.shape[0], y_true.shape[0]))
        return
    if x.shape[0] != y_pred.shape[0]:
        termwarn('Sample count mismatch: x(%d) != y_pred(%d). skipping evaluation' % (x.shape[0], y_pred.shape[0]))
        return

    y_pred_discrete = y_pred > 0.5

    images = [Image(img) for img in x]
    labels = [Image(mask) for mask in y_true]
    predictions = [Image(mask) for mask in y_pred]
    predictions_discrete = [Image(mask) for mask in y_pred_discrete]

    intersection = np.logical_and(y_pred_discrete, y_true)
    union = np.logical_or(y_pred_discrete, y_true)
    difference = np.logical_xor(y_pred_discrete, y_true)

    flat_shape = (x.shape[0], -1)

    accuracy = np.mean(np.equal(y_true, y_pred_discrete).reshape(flat_shape), axis=1)
    iou = np.sum(intersection.reshape(flat_shape), axis=1) / (np.sum(union.reshape(flat_shape), axis=1) + 1e-9)

    incorrect_predictions = [Image(mask) for mask in difference]

    if example_ids is None:
        example_ids = ['example_' + str(i) for i in range(len(x))]

    dfMap = {
        'wandb_example_id': example_ids,
        'image': images,
        'label': labels,
        'prediction': predictions,
        'prediction_discrete': predictions_discrete,
        'incorrect_prediction': incorrect_predictions,
        'accuracy': accuracy,
        'iou': iou,
    }

    all_columns = [
        'wandb_example_id',
        'image',
        'label',
        'prediction',
        'prediction_discrete',
        'incorrect_prediction',
        'accuracy',
        'iou',
    ]

    return pd.DataFrame(dfMap, columns=all_columns)

def image_segmentation_multiclass_dataframe(x, y_true, y_pred, labels, example_ids=None, class_colors=None):
    np = util.get_module('numpy', required='dataframes require numpy')
    pd = util.get_module('pandas', required='dataframes require pandas')

    x, y_true, y_pred= np.array(x), np.array(y_true), np.array(y_pred)

    if x.shape[0] != y_true.shape[0]:
        termwarn('Sample count mismatch: x(%d) != y_true(%d). skipping evaluation' % (x.shape[0], y_true.shape[0]))
        return
    if x.shape[0] != y_pred.shape[0]:
        termwarn('Sample count mismatch: x(%d) != y_pred(%d). skipping evaluation' % (x.shape[0], y_pred.shape[0]))
        return
    if class_colors is not None and len(class_colors) != y_true.shape[-1]:
        termwarn('Class color count mismatch: y_true(%d) != class_colors(%d). using generated colors' % (y_true.shape[-1], len(class_colors)))
        class_colors = None

    class_count = y_true.shape[-1]

    if class_colors is None:
        class_colors = util.class_colors(class_count)
    class_colors = np.array(class_colors)

    y_true_class = np.argmax(y_true, axis=-1)
    y_pred_class = np.argmax(y_pred, axis=-1)

    y_pred_discrete = np.round(y_pred)

    images = [Image(img) for img in x]
    label_imgs = [Image(mask) for mask in class_colors[y_true_class]]
    predictions = [Image(mask) for mask in class_colors[y_pred_class]]

    flat_shape = (x.shape[0], -1)

    intersection = np.sum(np.logical_and(y_true, y_pred_discrete).reshape(flat_shape), axis=1)
    union = np.sum(np.logical_or(y_true, y_pred_discrete).reshape(flat_shape), axis=1)

    iou = intersection / (union + 1e-9)
    accuracy = np.mean(np.equal(y_true_class, y_pred_class).reshape(flat_shape), axis=1)

    difference = np.zeros(y_true_class.shape)
    difference[y_true_class != y_pred_class] = 1.

    incorrect_predictions = [Image(mask) for mask in difference]

    iou_class = [
        np.sum(np.logical_and(y_true_class == i, y_pred_class == i).reshape(flat_shape), axis=1) / # intersection
        (np.sum(np.logical_or(y_true_class == i, y_pred_class == i).reshape(flat_shape), axis=1) + 1e-9) # union
        for i in range(len(labels))
    ]

    if example_ids is None:
        example_ids = ['example_' + str(i) for i in range(len(x))]

    dfMap = {
        'wandb_example_id': example_ids,
        'image': images,
        'label': label_imgs,
        'prediction': predictions,
        'incorrect_prediction': incorrect_predictions,
        'iou': iou,
        'accuracy': accuracy,
    }

    for i in range(len(iou_class)):
        dfMap['iou_{}'.format(labels[i])] = iou_class[i]

    all_columns = [
        'wandb_example_id',
        'image',
        'label',
        'prediction',
        'incorrect_prediction',
        'iou',
        'accuracy',
    ] + ['iou_'.format(l) for l in labels]

    return pd.DataFrame(dfMap, columns=all_columns)



import pytest
import numpy as np
import lightgbm as lgb
import wandb
import sys
from wandb import wandb_run
from wandb.lightgbm import plot_feature_importances, wandb_callback

from sklearn.linear_model import ElasticNet
from sklearn.datasets import make_classification

# Tests which rely on row history in memory should set `History.keep_rows = True`
from wandb.history import History

History.keep_rows = True


@pytest.fixture
def dummy_data(request):
    data = np.array([[1, 2], [3, 4]])
    label = np.array([0, 1])
    dtrain = lgb.Dataset(data, label=label)
    return dtrain


@pytest.mark.skipif(sys.version_info < (3, 6), reason="lightgbm was segfaulting in CI")
def test_basic_lightgbm(dummy_data, wandb_init_run):
    param = {'max_depth': 2, 'eta': 1}
    num_round = 2
    lgb.train(param, dummy_data, num_round, callbacks=[wandb_callback()])


def test_feature_importance_attribute_does_not_exists(wandb_init_run):
    data = np.array([[1, 2], [3, 4]])
    label = np.array([0, 1])
    model = ElasticNet(random_state=42)
    model.fit(data, label)
    dummy_features = []

    result = plot_feature_importances(model, feature_names=dummy_features)

    assert result is None, "Should have NOT returned any result, as feature_importances() attribute does NOT exist"


def test_feature_importance_attribute_exists_for_lightgbm(wandb_init_run):
    X, y = make_classification(n_samples=100, n_features=10, n_informative=2, n_redundant=5, random_state=42)
    ten_features = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']
    params = {'learning_rate': 0.01, 'max_depth': -1, 'num_leaves': 4, 'objective': 'fair', 'boosting': 'gbdt',
              'boost_from_average': True, 'feature_fraction': 0.9, 'bagging_freq': 1, 'bagging_fraction': 0.5,
              'early_stopping_rounds': 200, 'metric': 'rmse', 'max_bin': 255, 'n_jobs': -1, 'verbosity': -1,
              'bagging_seed': 1234}
    dataset = lgb.Dataset(X, y)
    model = lgb.train(params, dataset, valid_sets=[dataset], valid_names=['train'], callbacks=[wandb_callback()])

    result = plot_feature_importances(model, feature_names=ten_features)

    assert isinstance(result,
                      wandb.viz.Visualize), "Should have returned a result, as feature_importances() attribute does exist"



import torch
import torch.nn as nn
import torch.nn.functional as F
import wandb
import pytest
import json
import os
import sys
import time
from pprint import pprint
from torchvision import models
from torch.autograd import Variable
from pkg_resources import parse_version

# Tests which rely on row history in memory should set `History.keep_rows = True`
from wandb.history import History
History.keep_rows = True

# TODO: FLAKY SPECS sometimes these specs are timing out


def dummy_torch_tensor(size, requires_grad=True):
    if parse_version(torch.__version__) >= parse_version('0.4'):
        return torch.ones(size, requires_grad=requires_grad)
    else:
        return torch.autograd.Variable(torch.ones(size), requires_grad=requires_grad)


class DynamicModule(nn.Module):
    def __init__(self):
        super(DynamicModule, self).__init__()
        self.choices = nn.ModuleDict({
            'conv': nn.Conv2d(10, 10, 3),
            'pool': nn.MaxPool2d(3)
        })
        self.activations = nn.ModuleDict([
            ['lrelu', nn.LeakyReLU()],
            ['prelu', nn.PReLU()]
        ])

    def forward(self, x, choice, act):
        x = self.choices[choice](x)
        x = self.activations[act](x)
        return x


class Discrete(nn.Module):
    def __init__(self):
        super(Discrete, self).__init__()

    def forward(self, x):
        return nn.functional.softmax(x, dim=0)


class DiscreteModel(nn.Module):
    def __init__(self, num_outputs=2):
        super(DiscreteModel, self).__init__()
        self.linear1 = nn.Linear(1, 10)
        self.linear2 = nn.Linear(10, num_outputs)
        self.dist = Discrete()

    def forward(self, x):
        x = self.linear1(x)
        x = self.linear2(x)
        return self.dist(x)


class ParameterModule(nn.Module):
    def __init__(self):
        super(ParameterModule, self).__init__()
        self.params = nn.ParameterList(
            [nn.Parameter(torch.ones(10, 10)) for i in range(10)])
        self.otherparam = nn.Parameter(torch.Tensor(5))

    def forward(self, x):
        # ParameterList can act as an iterable, or be indexed using ints
        for i, p in enumerate(self.params):
            x = self.params[i // 2].mm(x) + p.mm(x)
        return x


def init_conv_weights(layer, weights_std=0.01, bias=0):
    '''Initialize weights for subnet convolution'''

    if parse_version(torch.__version__) >= parse_version('0.4'):
        nn.init.normal_(layer.weight.data, std=weights_std)
        nn.init.constant_(layer.bias.data, val=bias)
    else:
        nn.init.normal(layer.weight.data, std=weights_std)
        nn.init.constant(layer.bias.data, val=bias)
    return layer


def conv3x3(in_channels, out_channels, **kwargs):
    '''Return a 3x3 convolutional layer for SubNet'''
    layer = nn.Conv2d(in_channels, out_channels, kernel_size=3, **kwargs)
    layer = init_conv_weights(layer)

    return layer


class SubNet(nn.Module):
    def __init__(self, mode, anchors=9, classes=80, depth=4,
                 base_activation=F.relu,
                 output_activation=F.sigmoid):
        super(SubNet, self).__init__()
        self.anchors = anchors
        self.classes = classes
        self.depth = depth
        self.base_activation = base_activation
        self.output_activation = output_activation

        self.subnet_base = nn.ModuleList([conv3x3(256, 256, padding=1)
                                          for _ in range(depth)])

        if mode == 'boxes':
            self.subnet_output = conv3x3(256, 4 * self.anchors, padding=1)
        elif mode == 'classes':
            # add an extra dim for confidence
            self.subnet_output = conv3x3(
                256, (1 + self.classes) * self.anchors, padding=1)

        self._output_layer_init(self.subnet_output.bias.data)

    def _output_layer_init(self, tensor, pi=0.01):
        fill_constant = 4.59  # - np.log((1 - pi) / pi)

        return tensor.fill_(fill_constant)

    def forward(self, x):
        for layer in self.subnet_base:
            x = self.base_activation(layer(x))

        x = self.subnet_output(x)
        x = x.permute(0, 2, 3, 1).contiguous().view(x.size(0),
                                                    x.size(2) * x.size(3) * self.anchors, -1)

        return x


class ConvNet(nn.Module):
    def __init__(self):
        super(ConvNet, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)


class LSTMModel(torch.nn.Module):
    def __init__(self, embedding_dim, hidden_dim):
        super(LSTMModel, self).__init__()
        vocabLimit = 100
        self.hidden_dim = hidden_dim
        self.embeddings = nn.Embedding(vocabLimit + 1, embedding_dim)
        self.lstm = nn.LSTM(embedding_dim, hidden_dim)
        self.linearOut = nn.Linear(hidden_dim, 2)

    def forward(self, inputs, hidden):
        x = self.embeddings(inputs).view(len(inputs), 1, -1)
        lstm_out, lstm_h = self.lstm(x, hidden)
        x = lstm_out[-1]
        x = self.linearOut(x)
        x = F.log_softmax(x, dim=1)
        return x, lstm_h

    def init_hidden(self):
        return (
            Variable(torch.zeros(1, 1, self.hidden_dim)), Variable(torch.zeros(1, 1, self.hidden_dim)))


class Sequence(nn.Module):
    def __init__(self):
        super(Sequence, self).__init__()
        self.lstm1 = nn.LSTMCell(1, 51)
        self.lstm2 = nn.LSTMCell(51, 51)
        self.linear = nn.Linear(51, 1)

    def forward(self, input, future=0):
        outputs = []
        h_t = dummy_torch_tensor((input.size(0), 51))
        c_t = dummy_torch_tensor((input.size(0), 51))
        h_t2 = dummy_torch_tensor((input.size(0), 51))
        c_t2 = dummy_torch_tensor((input.size(0), 51))

        for i, input_t in enumerate(input.chunk(input.size(1), dim=1)):
            h_t, c_t = self.lstm1(input_t, (h_t, c_t))
            h_t2, c_t2 = self.lstm2(h_t, (h_t2, c_t2))
            output = self.linear(h_t2)
            outputs += [output]
        for i in range(future):  # if we should predict the future
            h_t, c_t = self.lstm1(output, (h_t, c_t))
            h_t2, c_t2 = self.lstm2(h_t, (h_t2, c_t2))
            output = self.linear(h_t2)
            outputs += [output]
        outputs = torch.stack(outputs, 1).squeeze(2)
        return outputs


class FCLayer(nn.Module):
    """FC Layer + Activation"""

    def __init__(self, dims, batchnorm_dim=0, act='ReLU', dropout=0):
        super(FCLayer, self).__init__()
        layers = []
        for i in range(len(dims) - 2):
            in_dim = dims[i]  # input
            out_dim = dims[i + 1]  # output

            if 0 < dropout:
                layers.append(nn.Dropout(dropout))
            print("BOOM", in_dim, out_dim)
            layers.append(nn.Linear(in_dim, out_dim))

            if '' != act:
                layers.append(getattr(nn, act)())
            if batchnorm_dim > 0:
                layers.append(nn.BatchNorm1d(batchnorm_dim))

        if 0 < dropout:
            layers.append(nn.Dropout(dropout))
        layers.append(nn.Linear(dims[-2], dims[-1]))

        if '' != act:
            layers.append(getattr(nn, act)())

        self.net = nn.Sequential(*layers)

    def forward(self, x):
        return self.net(x)


class VGGConcator(nn.Module):
    """
    Extracts feature of each four panels, concatenates 4 vgg features panel-wise.
    """

    def __init__(self):
        super(VGGConcator, self).__init__()
        self.vgg = models.vgg16(pretrained=False)
        self.vgg.classifier = nn.Sequential(
            *list(self.vgg.classifier.children())[:-1])

    def forward(self, panels, num=1):
        if num == 1:
            features = self.vgg(panels)
        else:
            img0 = panels[:, 0, :, :, :]
            img1 = panels[:, 1, :, :, :]
            img2 = panels[:, 2, :, :, :]
            img3 = panels[:, 3, :, :, :]

            feature0 = self.vgg(img0)
            feature1 = self.vgg(img1)
            feature2 = self.vgg(img2)
            feature3 = self.vgg(img3)

            features = torch.cat((feature0[:, None, :], feature1[:, None, :],
                                  feature2[:, None, :], feature3[:, None, :]), dim=1)

        return features


class Embedding(nn.Module):
    def __init__(self, d_embedding, d_word, d_hidden, word_dim, dropout, sparse=False):
        super(Embedding, self).__init__()

        glove = torch.ones((10, 300))
        self.vgg = VGGConcator()
        # self.fine_tuning()
        self.word_dim = word_dim
        glove = glove[:self.word_dim]

        self.d_word = d_word
        self.emb = nn.Embedding(word_dim, 300, padding_idx=0, sparse=sparse)
        self.emb.weight.data = glove

        # consts
        self.d_img = 4096
        self.num_panels = 4
        self.num_max_sentences = 3
        self.num_max_words = 20

        self.img_fc0 = FCLayer([self.d_img, d_hidden], dropout=dropout)

        self.box_lstm = nn.LSTM(
            300, d_word, 1, batch_first=True, bidirectional=False)
        self.fc0 = FCLayer([d_word + d_hidden, d_embedding])

    def forward(self, images, words):
        words = words.long()
        batch_size = words.size(0)
        box_hidden = self.init_hidden(batch_size, self.d_word, 1, 4)

        words = words.view(-1, words.size(-1))
        emb_word = self.emb(words)
        print("Shapes: ", emb_word.shape, words.shape)
        emb_word = emb_word.view(-1, self.num_panels,
                                 self.num_max_sentences, self.num_max_words, self.d_word)
        emb_sentence = torch.sum(emb_word, dim=3)
        emb_sentence = emb_sentence.view(-1,
                                         self.num_max_sentences, self.d_word)
        lstmed_sentence, _ = self.box_lstm(emb_sentence, box_hidden)
        emb_panel_sentence = lstmed_sentence[:, -1, :]

        emb_panel_sentence = emb_panel_sentence.view(
            -1, self.num_panels, self.d_word)

        img_feature = self.vgg(images, num=4)
        img_feature = self.img_fc0(img_feature)

        fusion = torch.cat((img_feature, emb_panel_sentence), dim=-1)
        fusion = self.fc0(fusion)
        return fusion

    def init_hidden(self, batch, out, direction=1, n=1):
        dims = (direction, batch * n, out)
        hiddens = (Variable(torch.zeros(*dims)),
                   Variable(torch.zeros(*dims)))
        return hiddens


@pytest.mark.skipif(sys.version_info < (3, 6), reason="Timeouts in older python versions")
def test_embedding(wandb_init_run):
    net = Embedding(d_embedding=300, d_word=300,
                    d_hidden=300, word_dim=100, dropout=0)
    wandb.watch(net, log="all", log_freq=1)
    for i in range(2):
        output = net(torch.ones((1, 4, 3, 224, 224)),
                     torch.ones((1, 4, 3, 20)))
        output.backward(torch.ones(1, 4, 300))
        wandb.log({"loss": 1})
    assert len(wandb_init_run.history.rows[0]) == 82


@pytest.mark.skipif(sys.version_info < (3, 6), reason="Timeouts in older python versions")
def test_sparse_embedding(wandb_init_run):
    net = Embedding(d_embedding=300, d_word=300,
                    d_hidden=300, word_dim=100, dropout=0, sparse=True)
    wandb.watch(net, log="all", log_freq=1)
    for i in range(2):
        output = net(torch.ones((1, 4, 3, 224, 224)),
                     torch.ones((1, 4, 3, 20)))
        output.backward(torch.ones(1, 4, 300))
        wandb.log({"loss": 1})
    assert len(wandb_init_run.history.rows[0]) == 82


def test_categorical(wandb_init_run):
    net = DiscreteModel(num_outputs=2)
    wandb.watch(net, log="all", log_freq=1)
    for i in range(2):
        output = net(torch.ones((1)))
        samp = output.backward(torch.ones((2)))
        wandb.log({"loss": samp})
    assert wandb_init_run.summary["graph_0"]._to_graph_json()
    assert len(wandb_init_run.history.rows[0]) == 12


def test_double_log(wandb_init_run):
    net = ConvNet()
    wandb.watch(net)
    with pytest.raises(ValueError):
        wandb.watch(net)


def test_gradient_logging(wandb_init_run):
    net = ConvNet()
    wandb.watch(net, log_freq=1)
    for i in range(3):
        output = net(dummy_torch_tensor((64, 1, 28, 28)))
        grads = torch.ones(64, 10)
        output.backward(grads)
        assert(len(wandb_init_run.history.row) == 8)
        assert(
            wandb_init_run.history.row['gradients/fc2.bias'].histogram[0] > 0)
        wandb.log({"a": 2})
    assert(len(wandb_init_run.history.rows) == 3)


def test_nan_inf_gradient_logging(wandb_init_run):
    net = ConvNet()
    wandb.watch(net, log_freq=1)

    output = net(dummy_torch_tensor((64, 1, 28, 28)))
    grads = torch.ones(64, 10)
    grads[32, 1] = float('nan')
    output.backward(grads)
    assert(len(wandb_init_run.history.row) == 6)
    assert('gradients/fc2.bias' not in wandb_init_run.history.row)
    wandb.log({"a": 2})

    output = net(dummy_torch_tensor((64, 1, 28, 28)))
    grads = torch.ones(64, 10)
    grads[0, 0] = float('inf')
    output.backward(grads)
    assert(len(wandb_init_run.history.row) == 6)
    assert('gradients/fc2.bias' not in wandb_init_run.history.row)
    wandb.log({"a": 2})

    assert(len(wandb_init_run.history.rows) == 2)


def test_unwatch(wandb_init_run):
    net = ConvNet()
    wandb.watch(net, log_freq=1, log="all")
    wandb.unwatch()
    for i in range(3):
        output = net(dummy_torch_tensor((64, 1, 28, 28)))
        grads = torch.ones(64, 10)
        output.backward(grads)
        assert(len(wandb_init_run.history.row) == 0)
        assert(
            wandb_init_run.history.row.get('gradients/fc2.bias') is None)
        wandb.log({"a": 2})
    assert(len(wandb_init_run.history.rows) == 3)


def test_unwatch_multi(wandb_init_run):
    net1 = ConvNet()
    net2 = ConvNet()
    wandb.watch(net1, log_freq=1, log="all")
    wandb.watch(net2, log_freq=1, log="all")
    wandb.unwatch(net1)
    for i in range(3):
        output1 = net1(dummy_torch_tensor((64, 1, 28, 28)))
        output2 = net2(dummy_torch_tensor((64, 1, 28, 28)))
        grads = torch.ones(64, 10)
        output1.backward(grads)
        output2.backward(grads)
        assert(len(wandb_init_run.history.row) == 16)
        print(wandb_init_run.history.row)
        assert wandb_init_run.history.row.get('gradients/graph_1conv1.bias')
        assert wandb_init_run.history.row.get('gradients/conv1.bias') is None
        wandb.log({"a": 2})
    assert(len(wandb_init_run.history.rows) == 3)


@pytest.mark.skipif(sys.version_info < (3, 6), reason="Timeouts in older python versions")
def test_gradient_logging_freq(wandb_init_run):
    net = ConvNet()
    log_freq = 50
    wandb.watch(net, log_freq=log_freq)
    for i in range(110):
        output = net(dummy_torch_tensor((64, 1, 28, 28)))
        grads = torch.ones(64, 10)
        output.backward(grads)
        if (i + 1) % log_freq == 0:
            assert(len(wandb_init_run.history.row) == 8)
            assert(
                wandb_init_run.history.row['gradients/fc2.bias'].histogram[0] > 0)
        else:
            assert(len(wandb_init_run.history.row) == 0)
        wandb.log({"a": 2})
    assert(len(wandb_init_run.history.rows) == 110)


def test_all_logging(wandb_init_run):
    net = ConvNet()
    wandb.watch(net, log="all", log_freq=1)
    for i in range(3):
        output = net(dummy_torch_tensor((64, 1, 28, 28)))
        grads = torch.ones(64, 10)
        output.backward(grads)
        assert(len(wandb_init_run.history.row) == 16)
        assert(
            wandb_init_run.history.row['parameters/fc2.bias'].histogram[0] > 0)
        assert(
            wandb_init_run.history.row['gradients/fc2.bias'].histogram[0] > 0)
        wandb.log({"a": 2})
    assert(len(wandb_init_run.history.rows) == 3)


@pytest.mark.skipif(sys.version_info < (3, 6), reason="Timeouts in older python versions")
def test_all_logging_freq(wandb_init_run):
    net = ConvNet()
    log_freq = 50
    wandb.watch(net, log="all", log_freq=log_freq)
    for i in range(110):
        output = net(dummy_torch_tensor((64, 1, 28, 28)))
        grads = torch.ones(64, 10)
        output.backward(grads)
        if (i + 1) % log_freq == 0:
            assert(len(wandb_init_run.history.row) == 16)
            assert(
                wandb_init_run.history.row['parameters/fc2.bias'].histogram[0] > 0)
            assert(
                wandb_init_run.history.row['gradients/fc2.bias'].histogram[0] > 0)
        else:
            assert(len(wandb_init_run.history.row) == 0)
        wandb.log({"a": 2})
    assert(len(wandb_init_run.history.rows) == 110)

# These were timing out in old python


@pytest.mark.skipif(sys.version_info < (3, 6), reason="Timeouts in older python versions")
def test_parameter_logging(wandb_init_run):
    net = ConvNet()
    wandb.watch(net, log="parameters", log_freq=1)
    for i in range(3):
        output = net(dummy_torch_tensor((64, 1, 28, 28)))
        grads = torch.ones(64, 10)
        output.backward(grads)
        assert(len(wandb_init_run.history.row) == 8)
        assert(
            wandb_init_run.history.row['parameters/fc2.bias'].histogram[0] > 0)
        wandb.log({"a": 2})
    assert wandb_init_run.summary["graph_0"]
    file_summary = json.loads(
        open(os.path.join(wandb_init_run.dir, "wandb-summary.json")).read())
    assert file_summary["graph_0"]
    assert(len(wandb_init_run.history.rows) == 3)


@pytest.mark.skipif(sys.version_info < (3, 6), reason="Timeouts in older python versions")
def test_parameter_logging_freq(wandb_init_run):
    net = ConvNet()
    log_freq = 20
    wandb.hook_torch(net, log="parameters", log_freq=log_freq)
    for i in range(50):
        #TO debug timeouts
        print("i: %i, time: %s" % (i, time.time()))
        output = net(dummy_torch_tensor((64, 1, 28, 28)))
        grads = torch.ones(64, 10)
        output.backward(grads)
        if (i + 1) % log_freq == 0:
            assert(len(wandb_init_run.history.row) == 8)
            assert(
                wandb_init_run.history.row['parameters/fc2.bias'].histogram[0] > 0)
        else:
            assert(len(wandb_init_run.history.row) == 0)
        wandb.log({"a": 2})
    assert(len(wandb_init_run.history.rows) == 50)


@pytest.mark.skipif(sys.version_info == (3, 6), reason="Timeouts in 3.6 for some reason...")
def test_simple_net():
    net = ConvNet()
    graph = wandb.wandb_torch.TorchGraph.hook_torch(net)
    output = net.forward(dummy_torch_tensor((64, 1, 28, 28)))
    grads = torch.ones(64, 10)
    output.backward(grads)
    graph = graph._to_graph_json()
    assert len(graph["nodes"]) == 5
    assert graph["nodes"][0]['class_name'] == "Conv2d(1, 10, kernel_size=(5, 5), stride=(1, 1))"
    assert graph["nodes"][0]['name'] == "conv1"


def test_sequence_net():
    net = Sequence()
    graph = wandb.wandb_torch.TorchGraph.hook_torch(net)
    output = net.forward(dummy_torch_tensor(
        (97, 100)))
    output.backward(torch.zeros((97, 100)))
    graph = graph._to_graph_json()
    assert len(graph["nodes"]) == 3
    assert len(graph["nodes"][0]['parameters']) == 4
    assert graph["nodes"][0]['class_name'] == "LSTMCell(1, 51)"
    assert graph["nodes"][0]['name'] == "lstm1"


def test_multi_net(wandb_init_run):
    net1 = ConvNet()
    net2 = ConvNet()
    graphs = wandb.watch((net1, net2))
    output1 = net1.forward(dummy_torch_tensor((64, 1, 28, 28)))
    output2 = net2.forward(dummy_torch_tensor((64, 1, 28, 28)))
    grads = torch.ones(64, 10)
    output1.backward(grads)
    output2.backward(grads)
    graph1 = graphs[0]._to_graph_json()
    graph2 = graphs[1]._to_graph_json()
    assert len(graph1["nodes"]) == 5
    assert len(graph2["nodes"]) == 5


def test_multi_net_global(wandb_init_run):
    net1 = ConvNet()
    net2 = ConvNet()
    wandb.watch(net1)
    wandb.watch(net2)
    output1 = net1.forward(dummy_torch_tensor((64, 1, 28, 28)))
    output2 = net2.forward(dummy_torch_tensor((64, 1, 28, 28)))
    grads = torch.ones(64, 10)
    output1.backward(grads)
    output2.backward(grads)
    assert wandb.run.summary["graph_1"]


def test_alex_net():
    alex = models.AlexNet()
    graph = wandb.wandb_torch.TorchGraph.hook_torch(alex)
    output = alex.forward(dummy_torch_tensor((2, 3, 224, 224)))
    grads = torch.ones(2, 1000)
    output.backward(grads)
    graph = graph._to_graph_json()
    # This was failing in CI with 21 nodes?!?
    print(graph["nodes"])
    assert len(graph["nodes"]) >= 20
    assert graph["nodes"][0]['class_name'] == "Conv2d(3, 64, kernel_size=(11, 11), stride=(4, 4), padding=(2, 2))"
    assert graph["nodes"][0]['name'] == "features.0"


def test_lstm(wandb_init_run):
    if parse_version(torch.__version__) < parse_version('0.4'):
        return

    net = LSTMModel(2, 2)
    graph = wandb.wandb_torch.TorchGraph.hook_torch(net)

    hidden = net.init_hidden()
    input_data = torch.ones((100)).type(torch.LongTensor)
    output = net.forward(input_data, hidden)
    grads = torch.ones(2, 1000)
    graph = graph._to_graph_json()

    assert len(graph["nodes"]) == 3
    assert graph["nodes"][2]['output_shape'] == [[1, 2]]


def test_resnet18():
    resnet = models.resnet18()
    graph = wandb.wandb_torch.TorchGraph.hook_torch(resnet)
    output = resnet.forward(dummy_torch_tensor((2, 3, 224, 224)))

    grads = torch.ones(2, 1000)
    output.backward(grads)
    graph = graph._to_graph_json()
    assert graph["nodes"][0]['class_name'] == "Conv2d(3, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False)"


def test_subnet():
    subnet = SubNet("boxes")
    graph = wandb.wandb_torch.TorchGraph.hook_torch(subnet)
    output = subnet.forward(dummy_torch_tensor((256, 256, 3, 3)))

    grads = torch.ones(256, 81, 4)
    output.backward(grads)
    graph = graph._to_graph_json()
    assert graph["nodes"][0]['class_name'] == "Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))"


def test_false_requires_grad(wandb_init_run):
    """When we set requires_grad to False, wandb must not
    add a hook to the variable"""

    net = ConvNet()
    net.fc1.weight.requires_grad = False
    wandb.watch(net, log_freq=1)
    output = net(dummy_torch_tensor((64, 1, 28, 28)))
    grads = torch.ones(64, 10)
    output.backward(grads)

    # 7 gradients are logged because fc1.weight is fixed
    assert(len(wandb_init_run.history.row) == 7)


def test_nested_shape():
    shape = wandb.wandb_torch.nested_shape([2, 4, 5])
    assert shape == [[], [], []]
    shape = wandb.wandb_torch.nested_shape([dummy_torch_tensor((2, 3)), dummy_torch_tensor((4, 5))])
    assert shape == [[2, 3], [4, 5]]
    # create recursive lists of tensors (t3 includes itself)
    t1 = dummy_torch_tensor((2, 3))
    t2 = dummy_torch_tensor((4, 5))
    t3 = [t1, t2]
    t3.append(t3)
    t3.append(t2)
    shape = wandb.wandb_torch.nested_shape([t1, t2, t3])
    assert shape == [[2, 3], [4, 5], [[2, 3], [4, 5], 0, [4, 5]]]
